# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""'(%)%now run' magics"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os
import sys
import argparse

from ...cmd.cmd_run import Run, run
from ...collection.metadata import Metascript, LAST_TRIAL
from ...persistence.models import Trial
from ...persistence import persistence_config

from .command import IpythonCommandMagic


class NowRun(IpythonCommandMagic, Run):
    """Run cells with noworkflow

    If it is in interactive mode (--interactive), it will use the context
    Otherwise, it will run in a subprocess, using the %%script magic.

    Please look at %%script? to understand --proc, --bg, --out and --err.

    Examples
    --------
    ::
        In [1]: my_var = 2

        In [2]: %%now_run --name script1 $my_var 30
           ...: import sys
           ...: print(sys.argv[1])
           ...: print(sys.argv[2])
        2
        30
        Out[2]: <Trial 1>

    ::
        In [3]: %%now_run --name script2 -v
           ...: l = range(4)
           ...: c = sum(l)
           ...: print(c)
        [now] removing noWorkflow boilerplate
        [now] setting up local provenance store
        [now] collecting definition provenance
        [now]   registering user-defined functions
        [now] collecting deployment provenance
        [now]   registering environment attributes
        [now]   searching for module dependencies
        [now]   registering provenance from 0 modules
        [now] collection execution provenance
        [now]   executing the script
        6
        [now] the execution of trial 2 finished successfully
        Out[3]: <Trial 2>

        In [4]: trial = _

    ::
        In [5]: %now_run script1.py $my_var 30
        Out[5]: <Trial 1>

    ::
        In [6]: trial = %now_run --name script2 script2.py
        6

    """

    def add_arguments(self):
        super(NowRun, self).add_arguments()
        add_arg = self.add_argument
        add_arg("--out", type=str,
                help="""The variable in which to store stdout from the script.
                If the script is backgrounded, this will be the stdout *pipe*,
                instead of the stdout text itself.
                """)
        add_arg("--err", type=str,
                help="""The variable in which to store stderr from the script.
                If the script is backgrounded, this will be the stderr *pipe*,
                instead of the stderr text itself.
                """)
        add_arg("--bg", action="store_true",
                help="""Whether to run the script in the background.
                If given, the only way to see the output of the command is
                with --out/err.
                """)
        add_arg("--proc", type=str,
                help="""The variable in which to store Popen instance.
                This is used only when --bg option is given.
                """)
        add_arg("--interactive", action="store_true",
                help="""Execute with the IPython context
                """)
        add_arg("params", nargs=argparse.REMAINDER,
                help="params to be passed to script")

    def execute(self, func, line, cell, magic_cls):                              # pylint: disable=too-many-locals
        # Calculate noworkflow params and script params
        argv, args = self.arguments(func, line)
        params = args.params
        if params:
            argv = argv[:-len(params)]
        # Create tmp file
        directory = args.dir or os.path.abspath(os.path.curdir)
        original_path = persistence_config.path
        try:
            persistence_config.connect(directory)

            if cell:
                filename = magic_cls.shell.mktempfile(data=cell,
                                                      prefix="now_run_")
            else:
                filename = params[0]
                params = params[1:]

            metascript = Metascript().read_ipython_args(
                args, directory, filename, [filename] + params,
                not args.interactive, cmd=line)

            if args.interactive:
                from IPython.utils.py3compat import builtin_mod
                save__name__ = magic_cls.shell.user_ns["__name__"]
                save__main__ = sys.modules["__main__"]
                save__argv__ = sys.argv
                save__path__ = sys.path[0]

                metascript.namespace = magic_cls.shell.user_ns

                metascript.clear_namespace(erase=False)
                metascript.clear_sys()
                sys.modules["__main__"] = magic_cls.shell.user_module
                run(metascript)

                magic_cls.shell.user_ns["__name__"] = save__name__
                magic_cls.shell.user_ns["__builtins__"] = builtin_mod
                sys.argv = save__argv__
                sys.path[0] = save__path__
                sys.modules["__main__"] = save__main__
                try:
                    return Trial(metascript.trial_id)
                except Exception as exc:                                         # pylint: disable=broad-except
                    print("Failed", exc)
            else:
                # Set execution line
                cmd = (
                    "now run --create_last {directory} {args} {script} "
                    "{params}"
                ).format(
                    directory="" if args.dir else "--dir {}".format(directory),
                    args=" ".join(argv),
                    script=filename,
                    params=" ".join(params)
                )
                script = magic_cls.shell.find_cell_magic("script").__self__
                script.shebang(cmd, "")
                tmp_dir = os.path.dirname(filename)

                try:
                    with open(os.path.join(tmp_dir, LAST_TRIAL), "r") as last:
                        return Trial(int(last.read()))
                except Exception as exc:                                         # pylint: disable=broad-except
                    print("Failed", exc)
        finally:
            persistence_config.connect(original_path)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Define executable bytecode interpreter"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import sys
import types
import dis

from dis import opmap

from collections import OrderedDict
from opcode import cmp_op

from .interpreter import Interpreter


CALL_FUNCTION = opmap["CALL_FUNCTION"]
CALL_FLAG_VAR = 1
CALL_FLAG_KW = 2


COMPARE = {
    cmp_op.index("<"): (lambda a, b: a < b),
    cmp_op.index("<="): (lambda a, b: a <= b),
    cmp_op.index("=="): (lambda a, b: a == b),
    cmp_op.index("!="): (lambda a, b: a != b),
    cmp_op.index(">"): (lambda a, b: a > b),
    cmp_op.index(">="): (lambda a, b: a >= b),
    cmp_op.index("in"): (lambda a, b: a in b),
    cmp_op.index("not in"): (lambda a, b: a not in b),
    cmp_op.index("is"): (lambda a, b: a is b),
    cmp_op.index("is not"): (lambda a, b: a is not b),
    cmp_op.index("exception match"): (lambda a, b: a == b),
    cmp_op.index("BAD"): (lambda a, b: False)
}


class CodeInterpreter(Interpreter):
    """Code Interpreter.
    Replace constructor to get code, locals and globals
    """

    def __init__(self, code, f_locals, f_globals, line_offset=0):
        super(CodeInterpreter, self).__init__(
            code.co_code, varnames=code.co_varnames, names=code.co_names,
            constants=code.co_consts,
            cells=code.co_cellvars + code.co_freevars,
            linestarts=OrderedDict(dis.findlinestarts(code)),
            line_offset=line_offset)

        self._code = code
        self._locals = f_locals
        self._globals = f_globals

        self.stack = []
        self.result = None

    def execute(self):
        """Iterate through interpreter"""
        for _ in self:
            pass

    def _pop_last_n(self, number):
        """Pop last n objects from stack"""
        return reversed([self.stack.pop() for _ in range(number)])


class ExecInterpreter(CodeInterpreter):                                          # pylint: disable=too-many-public-methods
    """Bytecode interpreter that executes it
    Currently, it is working only for Expressions
    """

    def __init__(self, code, f_locals, f_globals, line_offset=0):
        if not hasattr(self, "_known_missing"):
            self._known_missing = set()
        self._known_missing |= {
            "BREAK_LOOP", "CONTINUE_LOOP", "RETURN_VALUE", "YIELD_VALUE",
            "IMPORT_STAR", "POP_BLOCK", "POP_EXCEPT", "END_FINALLY",
            "SETUP_WITH", "WITH_CLEANUP_START", "WITH_CLEANUP_FINISH",
            "WITH_CLEANUP", "IMPORT_NAME", "IMPORT_FROM", "SETUP_LOOP",
            "SETUP_EXCEPT", "SETUP_FINALLY", "LOAD_CLOSURE", "LOAD_DEREF",
            "LOAD_CLASSDEREF", "STORE_DEREF", "DELETE_DEREF", "RAISE_VARARGS",
            "MAKE_CLOSURE", "UNPACK_SEQUENCE"
        }
        super(ExecInterpreter, self).__init__(
            code, f_locals, f_globals, line_offset=0)

    def _call(self, nargs, flags=0, nkw=0):
        """Execute call"""
        var, kwargs = [], {}
        if flags & CALL_FLAG_KW:
            kwargs = self.stack.pop()
        if flags & CALL_FLAG_VAR:
            var = self.stack.pop()
        for _ in range(nkw):
            value = self.stack.pop()
            key = self.stack.pop()
            kwargs[key] = value
        args = list(self._pop_last_n(nargs))
        func = self.stack[-1]
        self.stack[-1] = func(*(args + var), **kwargs)

    def _binary(self, func):
        """Execute binary operation"""
        tos = self.stack.pop()
        self.stack[-1] = func(self.stack[-1], tos)

    def _inplace(self, func):
        """Execute inplace operation"""
        tos = self.stack.pop()
        self.stack[-1] = func(self.stack[-1], tos)

    def _unary(self, func):
        """Execute unary operation"""
        self.stack[-1] = func(self.stack[-1])

    def extended_arg(self):
        """EXTENDED_ARG opcode"""
        self._extended_arg = self.oparg * 65536

    def load_fast(self):
        """LOAD_FAST opcode"""
        self.stack.append(self._locals[self.varnames[self.oparg]])

    def load_global(self):
        """LOAD_GLOBAL opcode"""
        self.stack.append(self._globals[self.names[self.oparg]])

    def load_attr(self):
        """LOAD_ATTR opcode"""
        self.stack[-1] = getattr(self.stack[-1], self.names[self.oparg])

    def load_const(self):
        """LOAD_CONST opcode"""
        self.stack.append(self.consts[self.oparg])

    def load_name(self):
        """LOAD_NAME opcode"""
        self.stack.append(self.names[self.oparg])

    def build_tuple(self):
        """BUILD_TUPLE opcode"""
        self.stack.append(tuple(self._pop_last_n(self.oparg)))

    def build_list(self):
        """BUILD_LIST opcode"""
        self.stack.append(list(self._pop_last_n(self.oparg)))

    def build_map(self):
        """BUILD_MAP opcode"""
        self.stack.append({})

    def build_set(self):
        """BUILD_SET opcode"""
        self.stack.append(set(self._pop_last_n(self.oparg)))

    def store_map(self):
        """STORE_MAP opcode"""
        key, value = self.stack.pop(), self.stack.pop()
        self.stack[-1][key] = value

    def call_function(self):
        """CALL_FUNCTION opcode"""
        self._call(self.oparg & 0xff, flags=(self.opcode - CALL_FUNCTION) & 3,
                   nkw=(self.oparg >> 8) & 0xff)

    def call_function_var(self):
        """CALL_FUNCTION_VAR opcode"""
        self.call_function()

    def call_function_kw(self):
        """CALL_FUNCTION_KW opcode"""
        self.call_function()

    def call_function_var_kw(self):
        """CALL_FUNCTION_VAR_KW opcode"""
        self.call_function()

    def make_function(self):
        """MAKE_FUNCTION opcode. Python 3 version"""
        qualname = self.stack.pop()
        func = types.FunctionType(self.stack.pop(), self._globals, qualname)
        posdefaults = self.oparg & 0xff
        kwdefaults = (self.oparg >> 8) & 0xff
        num_annotations = (self.oparg >> 16) & 0x7fff

        if num_annotations:
            _names = self.stack.pop()
            _values = self._pop_last_n(num_annotations - 1)
            func.__annotations__ = dict(zip(_names, _values))

        if kwdefaults:
            defaults = {}
            for _ in range(kwdefaults):
                _value = self.stack.pop()
                _key = self.stack.pop()
                defaults[_key] = _value
            func.__kwdefaults__ = defaults

        tup = tuple(self._pop_last_n(posdefaults))
        func.__defaults__ = tup

        self.stack.append(func)

    def build_slice(self):
        """BUILD SLICE opcode"""
        self.stack.append(slice(*self._pop_last_n(self.oparg)))

    def binary_add(self, func=lambda a, b: a + b):
        """BINARY_ADD opcode"""
        self._binary(func)

    def binary_and(self, func=lambda a, b: a & b):
        """BINARY_AND opcode"""
        self._binary(func)

    def binary_floor_divide(self, func=lambda a, b: a // b):
        """BINARY_FLOOR_DIVIDE opcode"""
        self._binary(func)

    def binary_lshift(self, func=lambda a, b: a << b):
        """BINARY_LSHIFT opcode"""
        self._binary(func)

    def binary_modulo(self, func=lambda a, b: a % b):
        """BINARY_MODULO opcode"""
        self._binary(func)

    def binary_multiply(self, func=lambda a, b: a * b):
        """BINARY_MULTIPLY opcode"""
        self._binary(func)

    def binary_or(self, func=lambda a, b: a | b):
        """BINARY_OR opcode"""
        self._binary(func)

    def binary_power(self, func=lambda a, b: a ** b):
        """BINARY_POWER opcode"""
        self._binary(func)

    def binary_rshift(self, func=lambda a, b: a >> b):
        """BINARY_RSHIFT opcode"""
        self._binary(func)

    def binary_subscr(self, func=lambda a, b: a[b]):
        """BINARY_SUBSCR opcode"""
        self._binary(func)

    def binary_subtract(self, func=lambda a, b: a - b):
        """BINARY_SUBTRACT opcode"""
        self._binary(func)

    def binary_true_divide(self, func=lambda a, b: a / b):
        """BINARY_TRUE_DIVIDE opcode"""
        self._binary(func)

    def binary_xor(self, func=lambda a, b: a ^ b):
        """BINARY_XOR opcode"""
        self._binary(func)

    def inplace_add(self, func=lambda a, b: a + b):
        """INPLACE_ADD opcode"""
        self._inplace(func)

    def inplace_and(self, func=lambda a, b: a & b):
        """INPLACE_AND opcode"""
        self._inplace(func)

    def inplace_floor_divide(self, func=lambda a, b: a // b):
        """INPLACE_FLOOR_DIVIDE opcode"""
        self._inplace(func)

    def inplace_lshift(self, func=lambda a, b: a << b):
        """INPLACE_LSHIFT opcode"""
        self._inplace(func)

    def inplace_modulo(self, func=lambda a, b: a % b):
        """INPLACE_MODULO opcode"""
        self._inplace(func)

    def inplace_multiply(self, func=lambda a, b: a * b):
        """INPLACE_MULTIPLY opcode"""
        self._inplace(func)

    def inplace_or(self, func=lambda a, b: a | b):
        """INPLACE_OR opcode"""
        self._inplace(func)

    def inplace_power(self, func=lambda a, b: a ** b):
        """INPLACE_POWER opcode"""
        self._inplace(func)

    def inplace_rshift(self, func=lambda a, b: a >> b):
        """INPLACE_RSHIFT opcode"""
        self._inplace(func)

    def inplace_subtract(self, func=lambda a, b: a - b):
        """INPLACE_SUBTRACT opcode"""
        self._inplace(func)

    def inplace_true_divide(self, func=lambda a, b: a / b):
        """INPLACE_TRUE_DIVIDE opcode"""
        self._inplace(func)

    def inplace_xor(self, func=lambda a, b: a ^ b):
        """INPLACE_XOR opcode"""
        self._inplace(func)

    def unary_invert(self, func=lambda a: ~a):
        """UNARY_INVERT opcode"""
        self._unary(func)

    def unary_negative(self, func=lambda a: -a):
        """UNARY_NEGATIVE opcode"""
        self._unary(func)

    def unary_not(self, func=lambda a: not a):
        """UNARY_NOT opcode"""
        self._unary(func)

    def unary_positive(self, func=lambda a: +a):
        """UNARY_POSITIVE opcode"""
        self._unary(func)

    def rot_two(self):
        """ROT_TWO opcode"""
        sta = self.stack
        sta[-1], sta[-2] = sta[-2], sta[-1]

    def rot_three(self):
        """ROT_THREE opcode"""
        sta = self.stack
        sta[-1], sta[-2], sta[-3] = sta[-2], sta[-3], sta[-1]

    def dup_top(self):
        """DUP_TOP opcode"""
        self.stack.append(self.stack[-1])

    def pop_top(self):
        """POP_TOP opcode"""
        self.stack.pop()

    def compare_op(self):
        """COMPARE_OP opcode"""
        tos, tos1 = self.stack.pop(), self.stack.pop()
        self.stack.append(COMPARE[self.oparg](tos1, tos))

    def jump_if_false_or_pop(self):
        """JUMP_IF_FALSE_OR_POP opcode"""
        tos = self.stack.pop()
        if not tos:
            self.lasti = self.oparg
            self.stack.append(tos)

    def jump_if_true_or_pop(self):
        """JUMP_IF_TRUE_OR_POP opcode"""
        tos = self.stack.pop()
        if tos:
            self.lasti = self.oparg
            self.stack.append(tos)

    def jump_forward(self):
        """JUMP_FORWARD opcode"""
        self.lasti += self.oparg

    def pop_jump_if_true(self):
        """POP_JUMP_IF_TRUE opcode"""
        if self.stack.pop():
            self.lasti = self.oparg

    def pop_jump_if_false(self):
        """POP_JUMP_IF_FALSE opcode"""
        if not self.stack.pop():
            self.lasti = self.oparg

    def jump_absolute(self):
        """JUMP_ABSOLUTE opcode"""
        self.lasti = self.oparg

    def get_iter(self):
        """GET_ITER opcode"""
        self.stack[-1] = iter(self.stack[-1])

    def list_append(self):
        """LIST_APPEND opcode"""
        tos = self.stack.pop()
        self.stack[-self.oparg].append(tos)

    def set_add(self):
        """SET_ADD opcode"""
        tos = self.stack.pop()
        self.stack[-self.oparg].add(tos)

    def map_add(self):
        """MAP_ADD opcode"""
        tos, tos1 = self.stack.pop(), self.stack.pop()
        self.stack[-self.oparg][tos] = tos1

    def for_iter(self):
        """FOR_ITER opcode"""
        try:
            self.stack.append(next(self.stack[-1]))
        except StopIteration:
            self.stack.pop()
            self.lasti += self.oparg

    def store_fast(self):
        """STORE_FAST opcode"""
        self._locals[self.varnames[self.oparg]] = self.stack.pop()

    def store_subscr(self):
        """STORE_SUBSCR opcode"""
        key, var, value = self.stack.pop(), self.stack.pop(), self.stack.pop()
        var[key] = value

    def store_name(self):
        """STORE_NAME opcode"""
        self._locals[self.names[self.oparg]] = self.stack.pop()

    def store_attr(self):
        """STORE_ATTR opcode"""
        var, value = self.stack.pop(), self.stack.pop()
        setattr(var, self.names[self.oparg], value)

    def store_global(self):
        """STORE_GLOBAL opcode"""
        self._globals[self.names[self.oparg]] = self.stack.pop()

    def delete_fast(self):
        """DELETE_FAST opcode"""
        del self._locals[self.varnames[self.oparg]]

    def delete_subscr(self):
        """DELETE_SUBSCR opcode"""
        key, var = self.stack.pop(), self.stack.pop()
        del var[key]

    def delete_name(self):
        """DELETE_NAME opcode"""
        del self._locals[self.names[self.oparg]]

    def delete_attr(self):
        """DELETE_ATTR opcode"""
        var = self.stack.pop()
        delattr(var, self.names[self.oparg])

    def delete_global(self):
        """DELETE_GLOBAL opcode"""
        del self._globals[self.names[self.oparg]]

    def print_expr(self):
        """PRINT_EXPR opcode"""
        self.result = self.stack.pop()


class Py2Codes(ExecInterpreter):
    """Bytecodes specific for Python 2"""

    def __init__(self, *args, **kwargs):
        if not hasattr(self, "_known_missing"):
            self._known_missing = set()
        self._known_missing |= {
            "BUILD_CLASS", "EXEC_STMT", "LOAD_LOCALS", "PRINT_ITEM",
            "PRINT_ITEM_TO", "PRINT_NEWLINE", "PRINT_NEWLINE_TO", "STOP_CODE"
        }
        super(Py2Codes, self).__init__(*args, **kwargs)

    def make_function(self):
        """MAKE_FUNCTION opcode"""
        tup = tuple(self._pop_last_n(self.oparg))
        func = types.FunctionType(self.stack.pop(), self._globals)
        func.func_defaults = tup
        self.stack.append(func)

    def binary_divide(self):
        """BINARY_DIVIDE opcode"""
        self._binary(lambda a, b: a / b)

    def binary_true_divide(self, func=lambda a, b: (
            float(a) if isinstance(a, int) else a) / b):
        """BINARY_TRUE_DIVIDE opcode"""
        self._binary(func)

    def inplace_divide(self):
        """INPLACE_DIVIDE opcode"""
        self._inplace(lambda a, b: a / b)

    def inplace_true_divide(self, func=lambda a, b: (
            float(a) if isinstance(a, int) else a) / b):
        """INPLACE_TRUE_DIVIDE opcode"""
        self._inplace(func)

    def unary_positive(self, func=repr):
        """UNARY_CONVERT opcode"""
        self._unary(func)

    def slice__0(self):
        """SLICE+0 opcode"""
        self.stack[-1] = self.stack[-1][:]

    def slice__1(self):
        """SLICE+1 opcode"""
        tos = self.stack.pop()
        self.stack[-1] = self.stack[-1][tos:]

    def slice__2(self):
        """SLICE+2 opcode"""
        tos = self.stack.pop()
        self.stack[-1] = self.stack[-1][:tos]

    def slice__3(self):
        """SLICE+3 opcode"""
        tos = self.stack.pop()
        tos1 = self.stack.pop()
        self.stack[-1] = self.stack[-1][tos1:tos]

    def rot_four(self):
        """ROT_FOUR opcode"""
        sta = self.stack
        sta[-1], sta[-2], sta[-3], sta[-4] = sta[-2], sta[-3], sta[-4], sta[-1]

    def dup_topx(self):
        """DUP_TOPX opcode"""
        topx = list(self._pop_last_n(self.oparg))
        self.stack = self.stack + topx + topx

    def store_slice__0(self):
        """STORE_SLICE+0 opcode"""
        var, value = self.stack.pop(), self.stack.pop()
        var[:] = value

    def store_slice__1(self):
        """STORE_SLICE+1 opcode"""
        sli, var, value = self.stack.pop(), self.stack.pop(), self.stack.pop()
        var[sli:] = value

    def store_slice__2(self):
        """STORE_SLICE+2 opcode"""
        sli, var, value = self.stack.pop(), self.stack.pop(), self.stack.pop()
        var[:sli] = value

    def store_slice__3(self):
        """STORE_SLICE+3 opcode"""
        sta = self.stack
        sli2, sli1, var, value = sta.pop(), sta.pop(), sta.pop(), sta.pop()
        var[sli1:sli2] = value

    def delete_slice__0(self):
        """DELETE_SLICE+0 opcode"""
        var = self.stack.pop()
        del var[:]

    def delete_slice__1(self):
        """STORE_SLICE+1 opcode"""
        sli, var = self.stack.pop(), self.stack.pop()
        del var[sli:]

    def delete_slice__2(self):
        """STORE_SLICE+2 opcode"""
        sli, var = self.stack.pop(), self.stack.pop()
        del var[:sli]

    def delete_slice__3(self):
        """STORE_SLICE+3 opcode"""
        sli2, sli1, var = self.stack.pop(), self.stack.pop(), self.stack.pop()
        del var[sli1:sli2]


class Py3Codes(ExecInterpreter):
    """Bytecodes specific for Python 3"""

    def __init__(self, *args, **kwargs):
        if not hasattr(self, "_known_missing"):
            self._known_missing = set()
        self._known_missing |= {
            "GET_YIELD_FROM_ITER", "BINARY_MATRIX_MULTIPLY", "GET_AWAITABLE",
            "GET_AITER", "GET_ANEXT", "BEFORE_ASYNC_WITH", "SETUP_ASYNC_WITH",
            "YIELD_FROM", "LOAD_BUILD_CLASS", "DELETE_DEREF",
            "LOAD_CLASSDEREF", "UNPACK_EX",
        }
        super(Py3Codes, self).__init__(*args, **kwargs)

    def dup_top_two(self):
        """DUP_TOP_TWO opcode"""
        top2 = list(self._pop_last_n(2))
        self.stack = self.stack + top2 + top2


PyInterpreter = (Py2Codes if sys.version_info >= (3, 0) else Py3Codes)           # pylint: disable=invalid-name



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Convert code the Jupyter Notebook files"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import json

def nbconvert(code):
    """Create Jupyter Notebook code


    Return dict in ipynb format
    Arguments:
    code -- code string separated by \\n
    """
    cells = []
    for cell in code.split("\n# <codecell>\n"):
        cells.append({
            "cell_type": "code",
            "execution_count": None,
            "metadata": {
                "collapsed": True,
            },
            "outputs": [],
            "source": [cell]
        })
    result = {
        "cells": cells,
        "nbformat": 4,
        "nbformat_minor": 0,
        "metadata": {
            "kernelspec": {
                "display_name": "Python 2",
                "language": "python",
                "name": "python2"
            },
            "language_info": {
                "codemirror_mode": {
                    "name": "ipython",
                    "version": 2
                },
                "file_extension": ".py",
                "mimetype": "text/x-python",
                "name": "python",
                "nbconvert_exporter": "python",
                "pygments_lexer": "ipython2",
                "version": "2.7.6"
            }
        }
    }
    return result


def create_ipynb(name, code):
    """Create ipynb file
    Store file with ipynb format


    Arguments:
    name -- filename
    code -- code string separated by \\n
    """
    inb = nbconvert(code)
    with open(name, "w") as ipynb:
        json.dump(inb, ipynb)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Define MetaProfiler

Use it to profile noWorkflow itself.
It is disabled by deault. To enabled, change active to True
"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import csv
import os

from collections import defaultdict
from datetime import datetime
from functools import wraps


class MetaProfiler(object):                                                      # pylint: disable=too-few-public-methods
    """Profile noWorkflow itself"""

    def __init__(self, active=False):
        self.file = "nowtime.csv"
        self.active = active
        self.order = [
            "cmd",
            "definition",
            "deployment", "environment", "modules",
            "execution",
            "storage"
        ]
        self.data = defaultdict(float)

    def __call__(self, typ):
        def dec(func):
            """Return decorator that stores the duration in typ"""
            if typ not in self.order:
                self.order.append(typ)

            @wraps(func)
            def wrapper(*args, **kwargs):
                """Capture duration of function"""
                before = datetime.now()
                result = func(*args, **kwargs)
                after = datetime.now()
                self.data[typ] += (after - before).total_seconds()
                return result
            return wrapper
        return dec

    def save(self):
        """Save durations"""
        if self.active:
            row = [self.data[name] for name in self.order]
            rows = []
            if not os.path.exists(self.file):
                rows.append(self.order)
            rows.append(row)

            with open(self.file, "a") as fil:
                writter = csv.writer(fil)
                writter.writerows(rows)

meta_profiler = MetaProfiler(active=False)                                       # pylint: disable=invalid-name



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Trial Model"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os

from sqlalchemy import Column, Integer, Text, TIMESTAMP
from sqlalchemy import ForeignKeyConstraint, select, func, distinct

from ...utils.formatter import PrettyLines
from ...utils.prolog import PrologDescription, PrologTrial, PrologNullableRepr
from ...utils.prolog import PrologTimestamp, PrologAttribute, PrologRepr
from ...utils.prolog import PrologNullable

from .. import relational, content, persistence_config

from .base import AlchemyProxy, proxy_class, query_many_property, proxy_gen
from .base import one, many_ref, many_viewonly_ref, backref_many, is_none
from .base import proxy

from .trial_prolog import TrialProlog
from .trial_dot import TrialDot

from .module import Module
from .dependency import Dependency
from .activation import Activation
from .head import Head
from .graphs.trial_graph import TrialGraph
from .graphs.dependency_graph import DependencyConfig, DependencyFilter
from .graphs.dependency_graph import PrologVisitor


@proxy_class                                                                     # pylint: disable=too-many-public-methods
class Trial(AlchemyProxy):
    """Represent a trial
    Initialize it by passing a trial reference:
        trial = Trial(2)


    There are four visualization modes for the graph:
        tree: activation tree without any filters
            trial.graph.mode = 0
        no match: tree transformed into a graph by the addition of sequence and
                  return edges and removal of intermediate call edges
            trial.graph.mode = 1
        exact match: calls are only combined when all the sub-call match
            trial.graph.mode = 2
        namesapce: calls are combined without considering the sub-calls
            trial.graph.mode = 3


    You can change the graph width and height by the variables:
        trial.graph.width = 600
        trial.graph.height = 400
    """

    __tablename__ = "trial"
    __table_args__ = (
        ForeignKeyConstraint(["inherited_id"], ["trial.id"],
                             ondelete="RESTRICT"),
        ForeignKeyConstraint(["parent_id"], ["trial.id"], ondelete="SET NULL"),
        {"sqlite_autoincrement": True},
    )

    id = Column(Integer, primary_key=True)                                       # pylint: disable=invalid-name
    start = Column(TIMESTAMP)
    finish = Column(TIMESTAMP)
    script = Column(Text)
    code_hash = Column(Text)
    arguments = Column(Text)
    command = Column(Text)
    inherited_id = Column(Integer, index=True)
    parent_id = Column(Integer, index=True)
    run = Column(Integer)
    docstring = Column(Text)

    inherited = one(
        "Trial", backref="bypass_children", viewonly=True,
        remote_side=[id], primaryjoin=(id == inherited_id)
    )
    parent = one(
        "Trial", backref="children", viewonly=True,
        remote_side=[id], primaryjoin=(id == parent_id)
    )

    function_defs = many_ref("trial", "FunctionDef")
    module_dependencies = many_ref("trials", "Dependency")
    dmodules = many_ref("trials", "Module", secondary=Dependency.t)
    environment_attrs = many_ref("trial", "EnvironmentAttr")
    activations = many_ref("trial", "Activation",
                           order_by=Activation.m.start)
    file_accesses = many_viewonly_ref("trial", "FileAccess")
    objects = many_viewonly_ref("trial", "Object")
    object_values = many_viewonly_ref("trial", "ObjectValue")
    variables = many_viewonly_ref("trial", "Variable")
    variable_usages = many_viewonly_ref("trial", "VariableUsage")
    variable_dependencies = many_viewonly_ref("trial", "VariableDependency")
    tags = many_ref("trial", "Tag")

    bypass_children = backref_many("bypass_children")  # Trial.inherited
    children = backref_many("children")  # Trial.parent

    @query_many_property
    def local_modules(self):
        """Load local modules. Return SQLAlchemy query"""
        return self.modules.filter(                                              # pylint: disable=no-member
            Module.m.path.like("%{}%".format(persistence_config.base_path)))

    @query_many_property
    def modules(self):
        """Load modules. Return SQLAlchemy query"""
        if self.inherited:
            return self.inherited.modules
        return self.dmodules

    @query_many_property
    def dependencies(self):
        """Load modules. Return SQLAlchemy query"""
        if self.inherited:
            return self.inherited.dependencies
        return self.module_dependencies

    @query_many_property
    def initial_activations(self):
        """Return initial activation as a SQLAlchemy query"""
        return self.activations.filter(is_none(Activation.m.caller_id))

    DEFAULT = {
        "dependency_config.show_blackbox_dependencies": False,
        "dot.format": "png",
        "graph.width": 500,
        "graph.height": 500,
        "graph.mode": 3,
        "graph.use_cache": True,
        "prolog.use_cache": True,
    }

    REPLACE = {
        "dependency_config_show_blackbox_dependencies":
            "dependency_config.show_blackbox_dependencies",
        "dot_format": "dot.format",
        "graph_width": "graph.width",
        "graph_height": "graph.height",
        "graph_mode": "graph.mode",
        "graph_use_cache": "graph.use_cache",
        "prolog_use_cache": "prolog.use_cache",
    }

    prolog_description = PrologDescription("trial", (
        PrologTrial("id"),
        PrologTimestamp("start"),
        PrologTimestamp("finish"),
        PrologRepr("script"),
        PrologRepr("code_hash"),
        PrologRepr("command"),
        PrologNullable("inherited_id", link="trial.id"),
        PrologNullable("parent_id", link="trial.id"),
        PrologAttribute("run"),
        PrologNullableRepr("docstring"),
    ), description=(
        "informs that a given *script* with *docstring*,\n"
        "and content *code_hash*,\n"
        "executed during a time period from *start*"
        "to *finish*,\n"
        "using noWokflow's *command*,\n"
        "that generated a trial *id*.\n"
        "This trial uses modules from *inherited_id*,\n"
        "is based on *parent_id*,\n"
        "and might be a *run* or a backup trial."
    ))

    def __init__(self, *args, **kwargs):
        if args and isinstance(args[0], relational.base):
            obj = args[0]
            trial_ref = obj.id
        elif args:
            trial_ref = kwargs.get("trial_ref", args[0])
        else:
            trial_ref = kwargs.get("trial_ref", None)

        # Check if it is a new trial or a query
        script = kwargs.get("trial_script", None)
        if "use_cache" in kwargs:
            cache = kwargs["use_cache"]
            kwargs["graph_use_cache"] = kwargs.get("graph_use_cache", cache)
            kwargs["prolog_use_cache"] = kwargs.get("graph_use_cache", cache)

        session = relational.session
        if not trial_ref or trial_ref == -1:
            obj = Trial.last_trial(script=script, session=session)
            if "graph_use_cache" not in kwargs:
                kwargs["graph_use_cache"] = False
            if "prolog_use_cache" not in kwargs:
                kwargs["prolog_use_cache"] = False
        else:
            obj = Trial.load_trial(trial_ref, session=session)

        if obj is None:
            raise RuntimeError("Trial {} not found".format(trial_ref))
        super(Trial, self).__init__(obj)
        #self._store_pk(obj)
        #self._restore_instance()

        self.dependency_config = DependencyConfig()
        self.dependency_filter = DependencyFilter(self)
        self.graph = TrialGraph(self)
        self.prolog = TrialProlog(self)
        self.dot = TrialDot(self)
        self.initialize_default(kwargs)
        self._prolog_visitor = None

    @property
    def prolog_variables(self):
        """Return filtered prolog variables"""
        if not self._prolog_visitor:
            self.dependency_filter.run()
            self._prolog_visitor = PrologVisitor(self.dependency_filter)
            self._prolog_visitor.visit(self.dependency_filter.main_cluster)
        return self._prolog_visitor

    @property
    def script_content(self):
        """Return the "main" script content of the trial"""
        return PrettyLines(
            content.get(self.code_hash)
            .decode("utf-8").split("/n"))

    @property
    def finished(self):
        """Check if trial has finished"""
        return bool(self.finish)

    @property
    def status(self):
        """Check trial status
        Possible statuses: finished, unfinished, backup"""
        if not self.run:
            return "backup"
        return "finished" if self.finished else "unfinished"

    @property
    def duration(self):
        """Calculate trial duration. Return microseconds"""
        if self.finish:
            return int((self.finish - self.start).total_seconds() * 1000000)
        return 0

    @property
    def duration_text(self):
        """Calculate trial duration. Return formatted str"""
        if self.finish:
            return str(self.finish - self.start)
        return "None"

    @property
    def environment(self):
        """Return dict: environment variables -> value"""
        return {e.name: e.value for e in self.environment_attrs}

    def versioned_files(self, skip_script=False, skip_local=False,
                        skip_access=False):
        """Find first files accessed in a trial
        Return map with relative path -> (code_hash, type)

        Possible types: script, module, access
        """
        files = {}
        def add(path, info):
            """Add file to dict"""
            if os.path.isabs(path):
                if not persistence_config.base_path in path:
                    return
                path = os.path.relpath(path, persistence_config.base_path)
            files[path] = info

        if not skip_script:
            add(self.script, {"code_hash": self.code_hash, "type": "script"})
        if not skip_local:
            for module in self.local_modules:                                    # pylint: disable=not-an-iterable
                add(module.path, {
                    "code_hash": module.code_hash,
                    "type": "module",
                    "name": module.name
                })
        if not skip_access:
            for faccess in reversed(list(self.file_accesses)):
                add(faccess.name, {
                    "code_hash": faccess.content_hash_before, "type": "access",
                })

        return files

    def iterate_accesses(self, path=None):
        """Iterate on all access to a path"""
        if not path or self.script.endswith(path):
            yield self.script, {"code_hash": self.code_hash, "type": "script"}
        for module in self.local_modules:                                        # pylint: disable=not-an-iterable
            if not path or module.path.endswith(path):
                yield module.path, {
                    "code_hash": module.code_hash,
                    "type": "module",
                    "name": module.name
                }
        for faccess in list(self.file_accesses):
            if not path or faccess.name.endswith(path):
                yield faccess.name, {
                    "code_hash": faccess.content_hash_before, "type": "access",
                }
                yield faccess.name, {
                    "code_hash": faccess.content_hash_after, "type": "access",
                }

    def create_head(self):
        """Create head for this trial"""
        session = relational.make_session()
        session.query(Head.m).filter(Head.m.script == self.script).delete()      # pylint: disable=no-member
        session.add(Head.m(trial_id=self.id, script=self.script))                # pylint: disable=no-member, not-callable
        session.commit()                                                         # pylint: disable=no-member

    def query(self, query):
        """Run prolog query"""
        return self.prolog.query(query)

    def _repr_html_(self):
        """Display d3 graph on ipython notebook"""
        if hasattr(self, "graph"):
            return self.graph._repr_html_()                                      # pylint: disable=protected-access
        return repr(self)

    def show(self, _print=print):
        """Print trial information"""
        _print("""\
            Id: {t.id}
            Inherited Id: {t.inherited_id}
            Script: {t.script}
            Code hash: {t.code_hash}
            Start: {t.start}
            Finish: {t.finish}
            Duration: {t.duration_text}\
            """.format(t=self))

    def __repr__(self):
        return "Trial({})".format(self.id)

    @classmethod  # query
    def distinct_scripts(cls):
        """Return a set with distinct scripts"""
        return {s[0].rsplit("/", 1)[-1]
                for s in relational.session.query(distinct(cls.m.script))}

    @classmethod  # query
    def reverse_trials(cls, limit, session=None):
        """Return a generator with <limit> trials ordered by start time desc"""
        session = session or relational.session
        return proxy_gen(
            session.query(cls.m)
            .order_by(cls.m.start.desc())
            .limit(limit)
        )

    @classmethod  # query
    def last_trial(cls, script=None, parent_required=False,
                   session=None):
        """Return last trial according to start time

        Keyword arguments:
        script -- specify the desired script (default=None)
        parent_required -- valid only if script exists (default=False)
        """
        model = cls.m
        session = session or relational.session
        trial = (
            session.query(model)
            .filter(model.start.in_(
                select([func.max(model.start)])
                .where(model.script == script)
            ))
        ).first()
        if trial or parent_required:
            return trial
        return (
            session.query(model)
            .filter(model.start.in_(
                select([func.max(model.start)])
            ))
        ).first()

    @classmethod  # query
    def find_by_name_and_time(cls, script, timestamp, trial=None,
                              session=None):
        """Return the first trial according to script and timestamp

        Arguments:
        script -- specify the desired script
        timestamp -- specify the start of finish time of trial

        Keyword Arguments:
        trial -- limit query to a specific trial
        """
        model = cls.m
        session = session or relational.session
        query = (
            session.query(model)
            .filter(
                (model.script == script) & (
                    model.start.like(timestamp + "%") |
                    model.finish.like(timestamp + "%")
                )
            ).order_by(model.start)
        )
        if trial:
            query = query.filter(model.id == trial)
        return proxy(query.first())


    @classmethod  # query
    def load_trial(cls, trial_ref, session=None):
        """Load trial by trial reference

        Find reference on trials id and tags name
        """
        from .tag import Tag  # avoid circular import
        session = session or relational.session
        return (
            session.query(cls.m)
            .outerjoin(Tag.m)
            .filter((cls.m.id == trial_ref) | (Tag.m.name == trial_ref))
        ).first()

    @classmethod  # query
    def load_parent(cls, script, remove=True, parent_required=False,
                    session=None):
        """Load head trial by script


        Keyword arguments:
        remove -- remove from head, after loading (default=True)
        parent_required -- valid only if script exists (default=False)
        session -- specify session for loading (default=relational.session)
        """
        session = session or relational.session
        head = Head.load_head(script, session=session)
        if head:
            trial = head.trial
            if remove:
                Head.remove(head.id, session=relational.make_session())
        elif not head:
            trial = cls.last_trial(
                script=script, parent_required=parent_required,
                session=session)
        return proxy(trial)

    @classmethod  # query
    def fast_last_trial_id(cls, session=None):
        """Load last trial id that did not bypass modules


        Compile SQLAlchemy core query into string for optimization

        Keyword arguments:
        session -- specify session for loading (default=relational.session)
        """
        session = session or relational.session
        if not hasattr(cls, "_last_trial_id"):
            ttrial = cls.t
            _query = (
                select([ttrial.c.id]).where(ttrial.c.start.in_(
                    select([func.max(ttrial.c.start)])
                    .select_from(ttrial)
                    .where(is_none(ttrial.c.inherited_id))
                ))
            )
            cls.last_trial_id = str(_query)
        an_id = session.execute(
            cls.last_trial_id).fetchone()
        if not an_id:
            raise RuntimeError(
                "Not able to bypass modules check because no previous trial "
                "was found"
            )
        return an_id[0]

    @classmethod  # query
    def fast_update(cls, trial_id, finish, docstring, session=None):
        """Update finish time of trial

        Use core sqlalchemy

        Arguments:
        trial_id -- trial id
        finish -- finish time as a datetime object


        Keyword arguments:
        session -- specify session for loading (default=relational.session)
        """
        session = session or relational.session
        ttrial = cls.t
        session.execute(
            ttrial.update()
            .values(finish=finish, docstring=docstring)
            .where(ttrial.c.id == trial_id)
        )
        session.commit()

    @classmethod  # query
    def store(cls, start, script, code_hash, arguments, bypass_modules,          # pylint: disable=too-many-arguments
              command, run, docstring, session=None):
        """Create trial and assign a new id to it

        Use core sqlalchemy

        Arguments:
        start -- trial start time
        script -- script name
        code_hash -- script hash code
        arguments -- trial arguments
        bypass_modules -- whether it captured modules or not
        command -- the full command line with noWorkflow parametes
        run -- trial created by the run command

        Keyword arguments:
        session -- specify session for loading (default=relational.session)
        """
        session = session or relational.session

        # ToDo: use core query
        parent = cls.load_parent(script, parent_required=True)
        parent_id = parent.id if parent else None

        inherited_id = None
        if bypass_modules:
            inherited_id = cls.fast_last_trial_id()
        ttrial = cls.__table__
        result = session.execute(
            ttrial.insert(),
            {"start": start, "script": script, "code_hash": code_hash,
             "arguments": arguments, "command": command, "run": run,
             "inherited_id": inherited_id, "parent_id": parent_id,
             "docstring": docstring})
        tid = result.lastrowid
        session.commit()
        return tid

    @classmethod  # query
    def all(cls, session=None):
        """Return all trials

        Keyword arguments:
        session -- specify session for loading (default=relational.session)
        """
        session = session or relational.session
        return proxy_gen(session.query(cls.m))



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Function Definition Model"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from sqlalchemy import Column, Integer, Text
from sqlalchemy import PrimaryKeyConstraint, ForeignKeyConstraint

from ...utils.prolog import PrologDescription, PrologTrial, PrologAttribute
from ...utils.prolog import PrologRepr, PrologNullableRepr

from .base import proxy_class, AlchemyProxy, many_ref, backref_one
from .base import query_many_property
from .object import Object


@proxy_class
class FunctionDef(AlchemyProxy):
    """Represent a function definition"""

    __tablename__ = "function_def"
    __table_args__ = (
        PrimaryKeyConstraint("trial_id", "id"),
        ForeignKeyConstraint(["trial_id"], ["trial.id"], ondelete="CASCADE"),
    )
    id = Column(Integer, index=True)                                             # pylint: disable=invalid-name
    name = Column(Text)
    code_hash = Column(Text)
    trial_id = Column(Integer, index=True)
    first_line = Column(Integer)
    last_line = Column(Integer)
    docstring = Column(Text)

    objects = many_ref("function_def", "Object")

    trial = backref_one("trial")  #  Trial.function_defs

    @query_many_property
    def globals(self):
        """Return function definition globals as a SQLAlchemy query"""
        return self.objects.filter(Object.m.type == "GLOBAL")

    @query_many_property
    def arguments(self):
        """Return function definition arguments as a SQLAlchemy query"""
        return self.objects.filter(Object.m.type == "ARGUMENT")

    @query_many_property
    def function_calls(self):
        """Return function definition calls as a SQLAlchemy query"""
        return self.objects.filter(Object.m.type == "FUNCTION_CALL")

    prolog_description = PrologDescription("function_def", (
        PrologTrial("trial_id", link="trial.id"),
        PrologAttribute("id"),
        PrologRepr("name"),
        PrologNullableRepr("code_hash"),
        PrologAttribute("first_line"),
        PrologAttribute("last_line"),
        PrologNullableRepr("docstring"),
    ), description=(
        "informs that in a given trial (*trial_id*),\n"
        "a function *name* was defined\n"
        "with content (*code_hash*)\n"
        "between *first_line* and *last_line*\n"
        "and with a *docstring*."
    ))

    def show(self, _print=lambda x, offset=0: print(x)):
        """Show object

        Keyword arguments:
        _print -- custom print function (default=print)
        """
        _print("""\
            Name: {f.name}
            Arguments: {arguments}
            Globals: {globals}
            Function calls: {calls}
            Code hash: {f.code_hash}\
            """.format(arguments=", ".join(x.name for x in self.arguments),      # pylint: disable=not-an-iterable
                       globals=", ".join(x.name for x in self.globals),          # pylint: disable=not-an-iterable
                       calls=", ".join(x.name for x in self.function_calls),     # pylint: disable=not-an-iterable
                       f=self))

    def __repr__(self):
        return "FunctionDef({0.trial_id}, {0.id}, {0.name})".format(self)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Define views for 'now vis'"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os

from flask import render_template, jsonify, request

from ..persistence.models import History, Diff, Trial
from ..persistence import relational


class WebServer(object):                                                         # pylint: disable=too-few-public-methods
    """Flask WebServer"""
    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(WebServer, cls).__new__(
                cls, *args, **kwargs)
        return cls._instance

    def __init__(self):
        from flask import Flask

        self.app = Flask(__name__)


app = WebServer().app                                                            # pylint: disable=invalid-name


@app.route("/<path:path>")
def static_proxy(path):
    """Serve static files"""
    return app.send_static_file(path)


@app.route("/trials")
def trials():
    """Respond history graph as JSON"""
    history = History(script=request.args.get("script"),
                      status=request.args.get("execution"))
    graph = history.graph.graph()
    return jsonify(**graph)


@app.route("/")
@app.route("/<tid>-<graph_mode>")
def index(tid=None, graph_mode=None):                                           # pylint: disable=unused-argument
    """Respond history scripts and index page as HTML"""
    history = History()
    return render_template(
        "index.html",
        cwd=os.getcwd(),
        scripts=history.scripts
    )


@app.route("/trials/<tid>/<graph_mode>/<cache>.json")
def trial_graph(tid, graph_mode, cache):
    """Respond trial graph as JSON"""
    trial = Trial(tid)
    graph = trial.graph
    graph.use_cache &= bool(int(cache))
    _, tgraph = getattr(graph, graph_mode)()
    return jsonify(**tgraph)


@app.route("/trials/<tid>/dependencies")
def dependencies(tid):
    """Respond trial module dependencies as JSON"""
    trial = Trial(tid)
    local = [x.to_dict() for x in trial.local_modules]                           # pylint: disable=not-an-iterable
    result = [x.to_dict() for x in trial.modules]                                # pylint: disable=not-an-iterable
    return jsonify(local=local, all=result)


@app.route("/trials/<tid>/all_modules")
def all_modules(tid):
    """Respond trial module dependencies as HTML"""
    trial = Trial(tid)
    local = trial.local_modules
    result = trial.modules
    result = sorted(result, key=lambda x: x not in local)                        # pylint: disable=unsupported-membership-test
    return render_template(
        "trial.html",
        cwd=os.getcwd(),
        trial=trial.to_dict(extra=("duration_text",)),
        modules=result,
        info="modules.html",
    )


@app.route("/trials/<tid>/environment")
def environment(tid):
    """Respond trial environment variables as JSON"""
    trial = Trial(tid)
    return jsonify(env=trial.environment)


@app.route("/trials/<tid>/all_environment")
def all_environment(tid):
    """Respond trial environment variables as HTML"""
    trial = Trial(tid)
    return render_template(
        "trial.html",
        cwd=os.getcwd(),
        trial=trial.to_dict(extra=("duration_text",)),
        env=trial.environment,
        info="environment.html",
    )


@app.route("/trials/<tid>/file_accesses")
def file_accesses(tid):
    """Respond trial file accesses as JSON"""
    trial = Trial(tid)
    return jsonify(file_accesses=[x.to_dict(extra=("stack",))
                                  for x in trial.file_accesses])


@app.route("/trials/<tid>/all_file_accesses")
def all_file_accesses(tid):
    """Respond trial file accesses as HTML"""
    trial = Trial(tid)
    return render_template(
        "trial.html",
        cwd=os.getcwd(),
        trial=trial.to_dict(extra=("duration_text",)),
        file_accesses=[x.to_dict(extra=("stack",))
                       for x in trial.file_accesses],
        info="file_accesses.html",
    )


@app.route("/diff/<trial1>/<trial2>")
@app.route("/diff/<trial1>/<trial2>/<tlimit>-<neighborhoods>-<graph_mode>")
def diff(trial1, trial2, tlimit=None, neighborhoods=None, graph_mode=None):      # pylint: disable=unused-argument
    """Respond trial diff as HTML"""
    diff_object = Diff(trial1, trial2)

    modules_added, modules_removed, modules_replaced = diff_object.modules
    env_added, env_removed, env_replaced = diff_object.environment
    fa_added, fa_removed, fa_replaced = diff_object.file_accesses
    return render_template(
        "diff.html",
        cwd=os.getcwd(),
        trial1=diff_object.trial1.to_dict(extra=("duration_text",)),
        trial2=diff_object.trial2.to_dict(extra=("duration_text",)),
        trial=diff_object.trial,
        modules_added=modules_added,
        modules_removed=modules_removed,
        modules_replaced=modules_replaced,
        env_added=env_added,
        env_removed=env_removed,
        env_replaced=env_replaced,
        fa_added=fa_added,
        fa_removed=fa_removed,
        fa_replaced=fa_replaced,
    )


@app.route("/diff/<trial1>/<trial2>/<graph_mode>/"
           "<tlimit>-<neighborhoods>-<cache>.json")                              # pylint: disable=too-many-arguments
def diff_graph(trial1, trial2, graph_mode, tlimit, neighborhoods, cache):
    """Respond trial diff as JSON"""
    diff_object = Diff(trial1, trial2)
    graph = diff_object.graph
    graph.use_cache &= bool(int(cache))

    _, diff_result, trial1, trial2 = getattr(graph, graph_mode)(
        time_limit=int(tlimit), neighborhoods=int(neighborhoods)
    )
    return jsonify(
        diff=diff_result,
        trial1=trial1,
        trial2=trial2,
    )


@app.teardown_appcontext
def shutdown_session(exception=None):                                            # pylint: disable=unused-argument
    """Shutdown SQLAlchemy session"""
    relational.session.remove()



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Helpers for AST analysis"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import ast
import sys
import itertools

from collections import namedtuple

from future.utils import viewvalues


CallDependency = namedtuple("Call", "line col")
ReturnDependency = namedtuple("Return", "line col")


class Variable(object):                                                          # pylint: disable=too-few-public-methods
    """Represent a variable name"""

    def __init__(self, name, typ):
        self.name = name
        self.type = typ

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if isinstance(other, str):
            return self.name == other
        elif isinstance(other, Variable):
            return self.name == other.name
        return False

    def __repr__(self):
        return "Var({}, {})".format(self.name, self.type)


class Dependency(object):                                                        # pylint: disable=too-few-public-methods
    """Represent a variable dependency"""

    def __init__(self, dependency, typ):
        self.dependency = dependency
        self.type = typ

    def __hash__(self):
        return hash(self.dependency)

    def __eq__(self, other):
        if not isinstance(other, Dependency):
            return False
        return self.dependency == other.dependency and self.type == other.type

    def __repr__(self):
        return "Dependency({}, {})".format(self.dependency, self.type)


def variable(name, typ):
    """Create Variable or Call Dependency"""
    if isinstance(name, str):
        return Variable(name, typ)
    if isinstance(name, (Variable, ReturnDependency, CallDependency)):
        return name
    if typ == "return":
        return ReturnDependency(*name)
    if typ in ("call", "print", "import", "import from"):
        return CallDependency(*name)
    return Variable(name, typ)


class NamedContext(object):
    """Store variable visibility context"""

    def __init__(self):
        self._names = [set()]
        self.use = False

    def flat(self):
        """Return available variable in the current context"""
        result = set()
        for name in self._names:
            result = result.union(name)
        return result

    def enable(self):
        """Enable variable collection"""
        self.use = True
        self._names.append(set())

    def disable(self):
        """Disable variable collection"""
        self.use = False

    def pop(self):
        """Remove sub-context from stack"""
        self._names.pop()

    def add(self, name):
        """Add variable to context"""
        if self.use:
            self._names[-1].add(name)


class DefinitionObject(object):

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, self.info())

    def _dependencies(self, node, visitor_class, func):                          # pylint: disable=no-self-use
        """Extract name dependencies from node"""
        visitor = visitor_class()
        visitor.visit(node)
        return [func(x if isinstance(x, FunctionCall) else x[0])
                for x in visitor.names]


class Loop(DefinitionObject):
    """Loop class. Used for For and While nodes"""

    def __init__(self, node, typ):
        self.node = node
        self.first_line = node.first_line
        self.last_line = node.last_line
        self.type = typ
        self.iterable = []
        self.iter_var = []
        self.maybe_call = None

    def info(self):
        """Return call information"""
        result = ("first_line={}, last_line={}")
        return result.format(self.first_line, self.last_line)

    def add_iterable(self, node, visitor_class):
        """Extract dependencies from iterable"""
        self.iterable = self._dependencies(
            node, visitor_class, lambda x: Dependency(x, "direct"))

    def add_iter_var(self, node, visitor_class):
        """Extract dependencies from iterable"""
        self.iter_var = self._dependencies(
            node, visitor_class, lambda x: Variable(x, "normal"))


class Condition(DefinitionObject):
    """Loop class. Used for If and While nodes"""

    def __init__(self, node):
        self.node = node
        self.first_line = node.first_line
        self.last_line = node.last_line
        self.test_var = []
        self.has_return = False

    def info(self):
        """Return call information"""
        result = ("first_line={}, last_line={}")
        return result.format(self.first_line, self.last_line)

    def add_test(self, node, visitor_class):
        """Extract dependencies from iterable"""
        self.test_var += self._dependencies(
            node, visitor_class, lambda x: Dependency(x, "conditional"))


class FunctionCall(ast.NodeVisitor):                                             # pylint: disable=too-many-instance-attributes
    """Represent a function call"""

    def __init__(self, visitor_class):
        self.self_attr = []
        self.func = []
        self.args = []
        self.keywords = {}
        self.starargs = []
        self.kwargs = []
        self.result = None
        self.line = -1
        self.col = -1
        self.visitor_class = visitor_class
        self.name = ""
        self.prefix = "call"

    def all_args(self):
        """List arguments of function call"""
        return [
            Dependency(x, "parameter")
            for x in itertools.chain(
                self.self_attr,
                itertools.chain.from_iterable(self.args),
                self.starargs,
                self.kwargs,
                itertools.chain.from_iterable(viewvalues(self.keywords))
            )
        ]

    def use_visitor(self, node):
        """Use configured visitor to visit sub node"""
        visitor = self.visitor_class()
        visitor.visit(node)
        return [x if isinstance(x, FunctionCall) else x[0]
                for x in visitor.names]

    def visit_Call(self, node):                                                  # pylint: disable=invalid-name
        """Visit Call"""
        self.func = self.use_visitor(node.func)
        if isinstance(node.func, ast.Attribute):
            self.self_attr = self.use_visitor(node.func.value)
        self.args = []
        for arg in node.args:
            if sys.version_info <= (3, 4) or not isinstance(arg, ast.Starred):
                self.args.append(self.use_visitor(arg))
            else:
                self.visit(arg)
        for keyword in node.keywords:
            self.visit(keyword)
        if hasattr(node, "starargs"):
            # Python <= 3.4
            if node.starargs:
                self.starargs = self.use_visitor(node.starargs)
            if node.kwargs:
                self.kwargs = self.use_visitor(node.kwargs)

    def visit_Starred(self, node):                                               # pylint: disable=invalid-name
        """Visit Starred. Only valid in Call context after Python 3.5"""
        # Python 3.5
        self.starargs += self.use_visitor(node)

    def visit_keyword(self, node):
        """Visit keyword"""
        if node.arg:
            self.keywords[node.arg] = self.use_visitor(node.value)
        else:
            # Python 3.5
            self.kwargs += self.use_visitor(node.value)

    def info(self):
        """Return call information"""
        result = ("line={}, col={}, "
                  "func={}, args={}, keywords={}, *args={}, **kwargs={}")
        return result.format(self.line, self.col, self.func, self.args,
                             self.keywords, self.starargs, self.kwargs)

    def __repr__(self):
        return "F({})".format(self.info())


class ClassDef(FunctionCall):
    """Represent a class definition"""

    def __repr__(self):
        return "Class(line={}, col={})".format(self.line, self.col)


class Decorator(FunctionCall):
    """Represent a decorator"""

    def __init__(self, *args, **kwargs):
        super(Decorator, self).__init__(*args, **kwargs)
        self.is_fn = True

    def __repr__(self):
        return "Decorator({})".format(self.info())

    def visit_Name(self, node):                                                  # pylint: disable=invalid-name
        """Visit Name"""
        self.func = self.use_visitor(node)
        self.is_fn = False

    def info(self):
        """Return decorator information"""
        if self.is_fn:
            return super(Decorator, self).info()
        return "line={}, col={}, name={}".format(
            self.line, self.col, self.func)


class Generator(FunctionCall):
    """Represent a generator"""

    def __init__(self, *args, **kwargs):
        self.type = args[-1]
        args = args[:-1]
        super(Generator, self).__init__(*args, **kwargs)

    def __repr__(self):
        return "Generator({})".format(self.info())

    def info(self):
        """Return generator information"""
        return "line={}, col={}, type={}".format(
            self.line, self.col, self.type)


class GeneratorCall(Generator):
    """Represent a generator call
    CALL_FUNCTION for set and dict comprehension on Python 2 and Python 3
    CALL_FUNCTION for list comprehension on Python 3
    """

    def __repr__(self):
        return "GeneratorCall({})".format(self.info())


class Assert(FunctionCall):
    """Represent an assert"""

    def __init__(self, *args, **kwargs):
        self.msg = args[-1]
        args = args[:-1]
        super(Assert, self).__init__(*args, **kwargs)

    def __repr__(self):
        return "Assert({})".format(self.info())

    def info(self):
        """Return assert information"""
        return "line={}, col={}, msg={}".format(
            self.line, self.col, self.msg)


class Print(FunctionCall):
    """Represent a print statement"""

    def __init__(self, *args, **kwargs):
        super(Print, self).__init__(*args, **kwargs)

    def __repr__(self):
        return "Print({})".format(self.info())


class With(FunctionCall):
    """Represent a with"""

    def __repr__(self):
        return "With({})".format(self.info())

    def info(self):
        """Return with information"""
        return "line={}, col={}".format(
            self.line, self.col)


class Import(FunctionCall):
    """Represent an import statement"""

    def __init__(self, *args, **kwargs):
        super(Import, self).__init__(*args, **kwargs)
        self.prefix = "import"

    def __repr__(self):
        return "Import(line={})".format(self.line)


class ForIter(FunctionCall):
    """Represent a for iter"""

    def __init__(self, *args, **kwargs):
        super(ForIter, self).__init__(*args, **kwargs)
        self.prefix = "iterator"

    def __repr__(self):
        return "ForIter({})".format(self.info())

    def info(self):
        """Return ForIter information"""
        return "line={}, col={}".format(
            self.line, self.col)


def index(lis, alternatives):
    """Return index of one of the <alternatives> in <lis>"""
    for alt in alternatives:
        try:
            return lis.index(alt)
        except ValueError:
            pass
    return None


def safeget(container, ind):
    """Try to access element in container. If it fails, prints container"""
    try:
        return container[ind]
    except IndexError as err:
        if not err.args:
            err.args = ("",)
        import pprint
        err.args = (err.args[0] + "\n Get\n  Index {}\n  Container \n{}".format(
            ind, pprint.pformat(list(enumerate(container)))),) + err.args[1:]
        raise err



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""History Model"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from .graphs.history_graph import HistoryGraph
from .base import Model
from .trial import Trial


class History(Model):
    """This model represents the workflow evolution history

    It is possible to filter the evolution history by selecting the script:
        history.script = "script1.py"

    The set of scripts can be accessed by:
        history.scripts

    It is also possible to filter the evolution history by selecting the
    trial status:
        history.status = "finished"

    The list of status are:
        finished: show only finished trials
        unfinished: show only unfinished trials
        backup: show only backup trials

    The default option for both filters is "*", which means that all trials
    appear in the history
        history.script = "*"
        history.status = "*"

    You can change the graph width and height by the variables:
        history.graph.width = 600
        history.graph.height = 200
    """
    __modelname__ = "History"

    DEFAULT = {
        "graph.width": 700,
        "graph.height": 300,
        "graph.use_cache": True,
        "script": "*",
        "status": "*",
    }

    REPLACE = {
        "graph_width": "graph.width",
        "graph_height": "graph.height",
        "graph_use_cache": "graph.use_cache",
    }

    def __init__(self, **kwargs):
        super(History, self).__init__(**kwargs)
        self.script = "*"
        self.status = "*"
        self.graph = HistoryGraph(self)
        self.initialize_default(kwargs)
        self.status_options = ["*", "finished", "unfinished", "backup"]

    @property
    def scripts(self):
        """Return a set of scripts used for trials"""
        return Trial.distinct_scripts()

    def _repr_html_(self):
        """Display d3 graph on ipython notebook"""
        return self.graph._repr_html_()                                          # pylint: disable=protected-access

    def __repr__(self):
        return repr(self.graph)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Environment Attribute Model"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from sqlalchemy import Column, Integer, Text
from sqlalchemy import PrimaryKeyConstraint, ForeignKeyConstraint

from ...utils.prolog import PrologDescription, PrologTrial, PrologRepr

from .base import AlchemyProxy, proxy_class, backref_one


@proxy_class
class EnvironmentAttr(AlchemyProxy):
    """Represent an environment attribute"""

    __tablename__ = "environment_attr"
    __table_args__ = (
        PrimaryKeyConstraint("trial_id", "id"),
        ForeignKeyConstraint(["trial_id"], ["trial.id"], ondelete="CASCADE"),
    )
    trial_id = Column(Integer, index=True)
    id = Column(Integer, index=True)                                             # pylint: disable=invalid-name
    name = Column(Text)
    value = Column(Text)

    trial = backref_one("trial")  # Trial.environment_attrs

    prolog_description = PrologDescription("environment", (
        PrologTrial("trial_id", link="trial.id"),
        PrologRepr("name"),
        PrologRepr("value"),
    ), description=(
        "informs that a environment attribute (*name*)\n"
        "was defined with *value*\n"
        "in a given trial (*trial_id*)."
    ))

    @property
    def brief(self):
        """Brief description of environment attribute"""
        return self.name

    def __hash__(self):
        return hash((self.name, self.value))

    def __eq__(self, other):
        return self.name == other.name

    def show(self, _print=lambda x, offset=0: print(x)):
        """Show object

        Keyword arguments:
        _print -- custom print function (default=print)
        """
        _print("{0.name}: {0.value}".format(self))

    def __repr__(self):
        return "Environment({0.trial_id}, {0.name}, {0.value})".format(self)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""'now restore' command"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import argparse
import os
import sys
import time

from future.utils import viewitems

from ..collection.metadata import Metascript
from ..persistence.models import Trial, Module, Dependency, FileAccess
from ..persistence import persistence_config, content
from ..utils.io import print_msg

from .command import Command


class Restore(Command):
    """Restore the files of a trial"""

    def __init__(self, *args, **kwargs):
        super(Restore, self).__init__(*args, **kwargs)
        self.print_msg = True

    def add_arguments(self):
        add_arg = self.add_argument
        add_arg("trial", type=str, nargs="?",
                help="trial id or none for last trial")
        add_arg("--dir", type=str,
                help="set project path where is the database. Default to "
                     "current directory")
        add_arg("-s", "--skip-script", action="store_true",
                help="skip main script")
        add_arg("-l", "--skip-local", action="store_true",
                help="skip local modules")
        add_arg("-a", "--skip-access", action="store_true",
                help="skip file access")

        add_arg("-f", "--file", nargs=argparse.REMAINDER,
                type=str,
                help="R|FORMAT: [-h] [-i ID] [-t TARGET] [path] \n"
                     "Restore a file from trial. Without <ID>, it restores \n"
                     "the first access to file. <ID> can be either the \n"
                     "timestamp, the number of access, or the code hash \n"
                     "<TARGET> specifies the target path (default to path)")


    def create_backup(self, metascript, files):
        """Create a backup trial"""
        modules = metascript.modules_store
        dependencies = metascript.dependencies_store
        accesses = metascript.file_accesses_store
        for path, info in viewitems(files):
            if info["type"] == "script":
                metascript.fake_path(path, "")
                metascript.paths[path].code_hash = info["code_hash"]
            elif info["type"] == "module":
                path = os.path.abspath(path)
                version = metascript.deployment.get_version(info["name"])
                mod = (info["name"], version, path, info["code_hash"])
                mid = Module.fast_load_module_id(*mod) or modules.add(*mod)
                dependencies.add(mid)
            elif info["type"] == "access":
                aid = accesses.add(path)
                access = accesses[aid]
                access.content_hash_before = info["code_hash"]
                access.content_hash_after = info["code_hash"]
                access.mode = "a+"

        metascript.trial_id = Trial.store(
            *metascript.create_trial_args(
                args="<restore {}>".format(metascript.trial_id), run=False
            )
        )

        tid, partial = metascript.trial_id, True
        Module.fast_store(tid, modules, partial)
        Dependency.fast_store(tid, dependencies, partial)
        FileAccess.fast_store(tid, accesses, partial)

        print_msg("Backup Trial {} created".format(metascript.trial_id),
                  self.print_msg)

    def restore(self, path, code_hash, trial_id, mode="normal"):
        """Restore file with <code_hash> from <trial_id>"""
        new_code_hash = file_hash(path)
        if code_hash == new_code_hash:
            return False
        if code_hash is None and mode != "script":
            os.remove(path)
            print_msg("File {} removed".format(path), self.print_msg)
            return True
        elif code_hash is not None:
            load_file = content.get(code_hash)
            with open(path, "wb") as fil:
                fil.write(load_file)
            msg = "File {}".format(path)
            if trial_id:
                msg += " from trial {}".format(trial_id)
            msg += " restored"

            print_msg(msg, self.print_msg)
            return True

    def restore_script(self, trial):
        """Restore the main script from <trial>"""
        self.restore(trial.script, trial.code_hash, trial.id)

    def find_differences(self, files):
        """Compare files with existing files
        Return match and dict with the most updated version of files

        match = True if all files match.
        """
        match = True
        new_files = {}

        for path, info in viewitems(files):
            new_code_hash = file_hash(path)
            if info["code_hash"] != new_code_hash:
                info["code_hash"] = new_code_hash
                match = False
                print_msg("{} has changed".format(path), self.print_msg)
            new_files[path] = info
        return match, new_files

    def execute(self, args):
        persistence_config.connect_existing(args.dir or os.getcwd())
        metascript = Metascript().read_restore_args(args)
        self.trial = trial = metascript.trial = Trial(trial_ref=args.trial)
        metascript.trial_id = trial.id
        metascript.name = trial.script
        metascript.fake_path(trial.script, "")
        metascript.paths[trial.script].code_hash = None

        restore_files = trial.versioned_files(**skip_dict(args))
        if not args.file:
            head = Trial.load_parent(trial.script, remove=False)

            last_files = head.versioned_files(**skip_dict(args))
            match, new_files = self.find_differences(last_files)
            if not match:
                self.create_backup(metascript, new_files)

            for path, info in viewitems(restore_files):
                self.restore(path, info["code_hash"], trial.id,
                             mode=info["type"])

            trial.create_head()
        else:
            parser = argparse.ArgumentParser(prog="{} restore {} file".format(
                os.path.basename(sys.argv[0]), trial.id))
            parser.add_argument("path", type=str)
            parser.add_argument(
                "-i", "--id", default="1", type=str,
                help="time ([X|]YYYY-MM-DD HH:MM:SS.ffffff), number of "
                     "access or hash. The optional X in the time, indicates "
                     "if we should get the content before of after the "
                     "access: X = B means before the access, X = A means "
                     "after. Example: 'A|2016-02-27 00:00' will query the "
                     "file accessed at midnight, and get its content after "
                     "the access")
            parser.add_argument(
                "-t", "--target", type=str,
                help="target path (default to path)")

            parsed = parser.parse_args(args.file)

            restore = self.find_file(args.trial, parsed.path, parsed.id)
            if not restore:
                print_msg("File version not found!", self.print_msg)
            else:
                path = parsed.target or parsed.path
                if not self.restore(path, restore[1], None):
                    print_msg("File has not changed!", self.print_msg)

    def find_file(self, tid, path, fid):                                         # pylint: disable=no-self-use
        """Find file according to timestamp, code_hash, and number of access"""
        splitted = fid.split("|")
        _time = time_str(splitted[-1])
        if _time:
            # timestamp
            text, _ = _time
            trial = Trial.find_by_name_and_time(path, text, trial=tid)
            if trial:
                return trial.script, trial.code_hash
            access = FileAccess.find_by_name_and_time(path, text, trial=tid)
            if access:
                if splitted[0] == "A":
                    return access.name, access.content_hash_after
                return access.name, access.content_hash_before
        if len(fid) >= 6:
            # at least 6 letters to look for code hash
            code_hash = content.find_subhash(fid)
            if code_hash:
                return path, code_hash

        if tid and fid.isnumeric():
            # Count accesses
            fid = int(fid)
            trial = Trial(tid)
            for i, result in enumerate(trial.iterate_accesses(path)):
                if fid == i + 1:
                    return path, result[1]["code_hash"]


def time_str(arg):
    """Convert arg to time with optional format.
    Return arg, converted time if converted or None, otherwise
    """
    current = ""
    extra = "%Y", "-%m", "-%d", " %H", ":%M", ":%S", ".%f"
    for part in extra:
        current += part
        try:
            converted = time.strptime(arg, current)
            return arg, converted
        except ValueError:
            pass
    return None


def file_hash(path):
    """Return hash of file in path or None"""
    abs_path = os.path.join(persistence_config.base_path, path)
    if not os.path.isfile(abs_path):
        return None
    else:
        with open(abs_path, "rb") as fil:
            return content.put(fil.read())


def skip_dict(args):
    """Skip specific files"""
    if args.file:
        return {}
    return {
        "skip_script": args.skip_script,
        "skip_local": args.skip_local,
        "skip_access": args.skip_access,
    }



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Slicing Dependency Model"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from sqlalchemy import Column, Integer, Text, select
from sqlalchemy import PrimaryKeyConstraint, ForeignKeyConstraint

from ...utils.prolog import PrologDescription, PrologTrial, PrologAttribute

from .. import relational

from .base import AlchemyProxy, proxy_class, backref_one


@proxy_class
class VariableDependency(AlchemyProxy):
    """Represent a variable dependency captured during program slicing"""

    __tablename__ = "variable_dependency"
    __table_args__ = (
        PrimaryKeyConstraint("trial_id", "id"),
        ForeignKeyConstraint(["trial_id"],
                             ["trial.id"], ondelete="CASCADE"),
        ForeignKeyConstraint(["trial_id", "source_activation_id"],
                             ["function_activation.trial_id",
                              "function_activation.id"], ondelete="CASCADE"),
        ForeignKeyConstraint(["trial_id", "target_activation_id"],
                             ["function_activation.trial_id",
                              "function_activation.id"], ondelete="CASCADE"),
        ForeignKeyConstraint(["trial_id",
                              "source_activation_id",
                              "source_id"],
                             ["variable.trial_id",
                              "variable.activation_id",
                              "variable.id"], ondelete="CASCADE"),
        ForeignKeyConstraint(["trial_id",
                              "target_activation_id",
                              "target_id"],
                             ["variable.trial_id",
                              "variable.activation_id",
                              "variable.id"], ondelete="CASCADE"),
    )
    trial_id = Column(Integer, index=True)
    id = Column(Integer, index=True)                                             # pylint: disable=invalid-name
    source_activation_id = Column(Integer, index=True)
    source_id = Column(Integer, index=True)
    target_activation_id = Column(Integer, index=True)
    target_id = Column(Integer, index=True)
    type = Column(Text)                                                          # pylint: disable=invalid-name

    trial = backref_one("trial")  # Trial.variable_dependencies
    # Activation.source_variables, Variable.dependencies_as_source
    source_activation = backref_one("source_activation")
    source = backref_one("source")
    # Activation.target_variables, Variable.dependencies_as_target
    target_activation = backref_one("target_activation")
    target = backref_one("target")

    prolog_description = PrologDescription("dependency", (
        PrologTrial("trial_id", link="variable.trial_id"),
        PrologAttribute("id"),
        PrologAttribute("source_activation_id", link="variable.activation_id"),
        PrologAttribute("source_id", link="variable.id"),
        PrologAttribute("target_activation_id", link="variable.activation_id"),
        PrologAttribute("target_id", link="variable.id"),
    ), description=(
        "informs that in a given trial (*trial_id*),\n"
        "the value of a variable (*target_id*)\n"
        "in a specific function activation (*target_activation_id*),\n"
        "influenced somehow the value of another variable (*source_id*)\n"
        "in another function activation (*source_activation_id*).\n"
        "This influence can occur due to direct assignment,\n"
        "matching of arguments in function activations,\n"
        "changes in mutable arguments of function activations,\n"
        "assignment within control flow structure, and function return."
    ))

    @classmethod  # query
    def fast_load_by_trial(cls, trial_id, session=None):
        """Return tuples with variable ids"""
        session = session or relational.session
        model = cls.m
        return session.execute(select([model.source_id, model.target_id])
            .where(model.trial_id == trial_id)
        )

    def __repr__(self):
        return (
            "VariableDependency({0.trial_id}, {0.id}, "
            "{0.source}, {0.target})"
        ).format(self)

    def __str__(self):
        return "{0.source} <- {0.target}".format(self)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Diff Graph Module"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import itertools
import time
import weakref

from collections import defaultdict, namedtuple
from copy import deepcopy, copy

from future.utils import viewitems, viewkeys, viewvalues

from ....utils.data import OrderedCounter, concat_iter, HashableDict

from .structures import prepare_cache, Graph


NeighborhoodConfig = namedtuple("NeighborhoodConfig", "check_caller getter")


def fix_caller_id(graph):
    """Fix IDs on graph"""
    called = {}
    seq = defaultdict(list)
    nodes = graph["nodes"]
    edges = [HashableDict(x) for x in graph["edges"]]
    for edge in edges:
        if edge["type"] == "call":
            called[edge["target"]] = edge["source"]
        if edge["type"] == "sequence":
            seq[edge["source"]].append(edge)

    visited = set()
    while called:
        new_called = {}
        for nid, parent in viewitems(called):
            nodes[nid]["caller_id"] = parent
            visited.add(nid)
            for edge in seq[nid]:
                if not edge["target"] in visited:
                    new_called[edge["target"]] = parent
        called = new_called


def prepare_graph(graph1, graph2):
    """Fix ids on graphs, match the <main> node
    Return hashable nodes1 and nodes2 from g1 and g2, respectively
    """
    fix_caller_id(graph1)
    fix_caller_id(graph2)
    nodes1 = [HashableDict(x) for x in graph1["nodes"]]
    nodes2 = [HashableDict(x) for x in graph2["nodes"]]
    graph1["hnodes"], graph2["hnodes"] = nodes1, nodes2
    graph1["node_indexes"] = set(range(len(nodes1)))
    graph2["node_indexes"] = set(range(len(nodes2)))
    graph1["levels"] = defaultdict(set)
    graph2["levels"] = defaultdict(set)
    for node in nodes1:
        graph1["levels"][node["node"]["level"]].add(node["index"])
    for node in nodes2:
        graph2["levels"][node["node"]["level"]].add(node["index"])
    if nodes1 and nodes2:
        if nodes1[0]['name'] != nodes2[0]['name']:
            nodes1[0]["name"] = "<main>"
            nodes2[0]["name"] = "<main>"
        graph1["max_level"] = max(viewkeys(graph1["levels"]))
        graph2["max_level"] = max(viewkeys(graph2["levels"]))
    else:
        graph1["max_level"] = -1
        graph2["max_level"] = -1

    return nodes1, nodes2


def lcs(lis1, lis2, equals=lambda x, y: x == y):
    """Longest common subsequence for generic lists lis1, lis2
    Return two OrderedCounters representing the matches
    """
    lengths = [[0 for _ in range(len(lis2) + 1)] for _ in range(len(lis1) + 1)]
    # row 0 and column 0 are initialized to 0 already
    for in1, element1 in enumerate(lis1):
        for in2, element2 in enumerate(lis2):
            if equals(element1, element2):
                lengths[in1 + 1][in2 + 1] = lengths[in1][in2] + 1
            else:
                lengths[in1 + 1][in2 + 1] = \
                    max(lengths[in1 + 1][in2], lengths[in1][in2 + 1])
    # read the substring out from the matrix
    matches1, matches2 = OrderedCounter(), OrderedCounter()
    len1, len2 = len(lis1), len(lis2)
    while len1 != 0 and len2 != 0:
        if lengths[len1][len2] == lengths[len1 - 1][len2]:
            len1 -= 1
        elif lengths[len1][len2] == lengths[len1][len2 - 1]:
            len2 -= 1
        else:
            matches1[lis1[len1 - 1]] = lis2[len2 - 1]
            matches2[lis2[len2 - 1]] = lis1[len1 - 1]
            len1 -= 1
            len2 -= 1
    return matches1, matches2


def cmp_node_fn(nodes1, nodes2):
    """Return a node comparing function for the two graphs"""
    def cmp_node(node1, node2, check_caller=True):
        """Compare two nodes"""
        if node1["name"] != node2["name"]:
            return False
        if check_caller:
            if node1["caller_id"] is None and node2["caller_id"] is not None:
                return False
            if node1["caller_id"] is not None and node2["caller_id"] is None:
                return False
            if node1["caller_id"] is None and node2["caller_id"] is None:
                return True
            caller1 = nodes1[node1["caller_id"]]
            caller2 = nodes2[node2["caller_id"]]
            return caller1["name"] == caller2["name"]
        return True
    return cmp_node


def cmp_edge_fn(nodes1, nodes2):
    """Return a edge comparing function for the two graphs"""
    cmp_node = cmp_node_fn(nodes1, nodes2)

    def cmp_edge(node1, node2):
        """Compare two edges"""
        if node1["type"] != node2["type"]:
            return False
        if not cmp_node(nodes1[node1["source"]], nodes2[node2["source"]]):
            return False
        if not cmp_node(nodes1[node1["target"]], nodes2[node2["target"]]):
            return False
        return True
    return cmp_edge


def first_solution(nodes1, nodes2):
    """Use LCS and return an initial mapping between nodes"""
    mapping, _ = lcs(nodes1, nodes2, cmp_node_fn(nodes1, nodes2))
    return mapping


class MappingToGraph(object):
    """Convert match to graph"""

    def __init__(self, g1, g2, mapping):
        self.nid = 0
        self._result = {
            "max_duration": dict(concat_iter(
                viewitems(g1["max_duration"]), viewitems(g2["max_duration"]))),
            "min_duration": dict(concat_iter(
                viewitems(g1["min_duration"]), viewitems(g2["min_duration"]))),
            "nodes": [],
            "edges": [],
        }
        self.context_edges = {}
        self.old_to_new = {}

        self.nodes1, self.nodes2 = g1["hnodes"], g2["hnodes"]
        del g1["hnodes"], g2["hnodes"]
        del g1["node_indexes"], g2["node_indexes"]
        del g1["levels"], g2["levels"]
        del g1["max_level"], g2["max_level"]
        self.merge(g1, g2, mapping)

    @property
    def nodes(self):
        """Return list of nodes"""
        return self._result["nodes"]

    @property
    def edges(self):
        """Return list of edges"""
        return self._result["edges"]

    def merge(self, graph1, graph2, mapping):
        """Merge matched graphs"""
        for node1, node2 in viewitems(mapping):
            cnode = deepcopy(node1)
            del cnode["node"]
            cnode["node1"] = node1["node"]
            cnode["node2"] = node2["node"]
            cnode["node1"]["original"] = node1["index"]
            cnode["node2"]["original"] = node2["index"]
            cnode["index"] = self.nid
            node1["node"]["diff"] = self.nid
            node2["node"]["diff"] = self.nid
            self.old_to_new[(1, node1["index"])] = self.nid
            self.old_to_new[(2, node2["index"])] = self.nid
            self.nid += 1
            self.nodes.append(cnode)

        self.add_nodes(self.nodes1, 1)
        self.add_nodes(self.nodes2, 2)
        self.add_edges(graph1["edges"], 1)
        self.add_edges(graph2["edges"], 2)

    def add_nodes(self, nodes, trial_number):
        """Convert matched nodes to graph"""
        for node in nodes:
            if (trial_number, node["index"]) not in self.old_to_new:
                nid = self.add_node(node)
                self.old_to_new[(trial_number, node["index"])] = nid
                node["node"]["diff"] = nid

    def add_edges(self, edges, trial_number):
        """Convert matched edges to graph"""
        for edge in edges:
            if (trial_number, edge["source"]) in self.old_to_new:
                self.add_edge(self.old_to_new[(trial_number, edge["source"])],
                              self.old_to_new[(trial_number, edge["target"])],
                              edge, trial_number)

    def add_node(self, node):
        """Convert matched node to graph"""
        cnode = deepcopy(node)
        node_id = cnode["index"] = self.nid
        cnode["node"]["original"] = node["index"]
        self.nodes.append(cnode)
        self.nid += 1
        return node_id

    def add_edge(self, source, target, edge, trial_number):
        """Convert matched edge to graph"""
        edge_key = "{} {} {}".format(source, target, edge["type"])

        if edge_key not in self.context_edges:
            cedge = deepcopy(edge)
            cedge["source"] = source
            cedge["target"] = target
            cedge["trial"] = trial_number
            self.edges.append(cedge)
            self.context_edges[edge_key] = cedge
        else:
            cedge = self.context_edges[edge_key]
            cedge["count"] = (cedge["count"], edge["count"])
            cedge["trial"] = 0

    def to_dict(self):
        """Create resulting dict"""
        return self._result


class Similarity(object):
    """Calculate similarity of a mapping between two graphs"""

    def __init__(self, g1, g2):
        self.graph1, self.graph2 = g1, g2
        self.total_nodes = float(self.count_union_nodes()) or 1.0
        self.total_edges = float(self.count_union_edges()) or 1.0
        self.edges1, self.edges2 = defaultdict(set), defaultdict(set)
        for edge in g1["edges"]:
            self.edges1[edge["source"]].add((edge["target"], edge["type"]))
        for edge in g2["edges"]:
            self.edges2[edge["source"]].add((edge["target"], edge["type"]))

    def count_union_nodes(self):
        """Count |A v B| where A and B are sets of nodes from the graphs"""
        nodes1, nodes2 = self.graph1["hnodes"], self.graph2["hnodes"]
        if len(nodes2) < len(nodes1):
            nodes2, nodes1 = nodes1, nodes2
        cmp_node = cmp_node_fn(nodes1, nodes2)

        used = [False] * len(nodes1)
        for node2 in nodes2:
            for i, node1 in enumerate(nodes1):
                if not used[i] and cmp_node(node1, node2):
                    used[i] = node2
                    break
        return sum(1 for i, _ in enumerate(used) if not used[i]) + len(nodes2)

    def count_union_edges(self):
        """Count |A v B| where A and B are sets of edges from the graphs"""
        nodes1, nodes2 = self.graph1["hnodes"], self.graph2["hnodes"]
        edges1, edges2 = self.graph1["edges"], self.graph2["edges"]
        if len(edges2) < len(edges1):
            edges2, edges1 = edges1, edges2
            nodes2, nodes1 = nodes1, nodes2
        cmp_edge = cmp_edge_fn(nodes1, nodes2)

        used = [False] * len(edges1)
        for edge2 in edges2:
            for i, edge1 in enumerate(edges1):
                if not used[i] and cmp_edge(edge1, edge2):
                    used[i] = edge2
                    break
        return sum(1 for i, _ in enumerate(used) if not used[i]) + len(edges2)

    def edge_intersection(self, mapping):
        """Count |A ^ B| where A and B are sets of edges from the graphs"""
        nodes1 = self.graph1["hnodes"]
        edges = 0
        for node1, node2 in viewitems(mapping):
            targets1 = self.edges1[node1["index"]]
            targets2 = self.edges2[node2["index"]]
            for target, typ in targets1:
                target1 = nodes1[target]
                if target1 not in mapping:
                    continue
                if (mapping[target1]["index"], typ) in targets2:
                    edges += 1
        return float(edges)

    def __call__(self, mapping):
        node_sim = (float(len(mapping)) / self.total_nodes)
        edge_sim = self.edge_intersection(mapping) / self.total_edges
        result = node_sim + edge_sim
        return result


def greedy(graph1, graph2):
    """Calculate graph matching through greedy algorithm"""
    nodes1, nodes2 = prepare_graph(graph1, graph2)
    mapping = first_solution(nodes1, nodes2)
    return MappingToGraph(graph1, graph2, mapping).to_dict()


def neighborhood1(graph1, graph2, mapping, cmp_node):
    """First neighborhood of VND. Add missing combinations"""
    tried = set()

    def add_to_mapping(to_add, new_mapping, swapped):
        """Add combination to mapping"""
        nodes1, nodes2 = graph1["hnodes"], graph2["hnodes"]
        added = []
        for node_id1, node_id2 in to_add:
            if swapped:
                node_id1, node_id2 = node_id2, node_id1
            node1, node2 = nodes1[node_id1], nodes2[node_id2]
            if cmp_node(node1, node2):
                added.append((node_id1, node_id2))
                new_mapping[node1] = node2
        return tuple(added)

    not_mapped1 = (graph1["node_indexes"] -
                   {n["index"] for n in viewkeys(mapping)})
    not_mapped2 = (graph2["node_indexes"] -
                   {n["index"] for n in viewvalues(mapping)})
    swapped = False
    if len(not_mapped2) > len(not_mapped1):
        swapped = True
        not_mapped1, not_mapped2 = not_mapped2, not_mapped1

    possibilities = [
        list(zip(x, not_mapped2))
        for x in itertools.permutations(not_mapped1, len(not_mapped2))]

    for full_map in possibilities:
        for i in range(1, len(not_mapped2) + 1):
            for to_add in itertools.combinations(full_map, i):
                if to_add in tried:
                    continue
                new_mapping = copy(mapping)
                to_add = add_to_mapping(to_add, new_mapping, swapped)
                if to_add in tried:
                    continue
                tried.add(to_add)
                yield new_mapping


def neighborhood2or3(graph1, graph2, mapping, cmp_node, config):
    """Second and Third neighborhoods of VND. Permutate matches"""
    nodes1, nodes2 = graph1["hnodes"], graph2["hnodes"]
    max_level = min(graph1["max_level"], graph2["max_level"])
    reverse = {n2: n1 for n1, n2 in viewitems(mapping)}

    def permutate(group1, group2, check_caller):
        """Permutate results in mapping"""
        for nid1, nid2 in itertools.product(group1, group2):
            node1, node2 = nodes1[nid1], nodes2[nid2]
            if id(mapping[node1]) == id(node2):
                # already on mapping
                continue
            if not cmp_node(node1, node2, check_caller):
                # different nodes
                continue
            new_mapping = copy(mapping)
            if node2 in reverse:
                del new_mapping[reverse[node2]]
            new_mapping[node1] = node2
            for match in neighborhood1(graph1, graph2, new_mapping, cmp_node):
                yield match

    for i in range(1, max_level + 1):
        param1 = config.getter(graph1, i)
        param2 = config.getter(graph2, i)
        for match in permutate(param1, param2, config.check_caller):
            yield match


def neighborhood(k, graph1, graph2, mapping):
    """Match nodes in the same neighborhood for VND"""
    nodes1, nodes2 = graph1["hnodes"], graph2["hnodes"]
    cmp_node = cmp_node_fn(nodes1, nodes2)

    if k == 1:
        generator = neighborhood1(graph1, graph2, mapping, cmp_node)

    if k == 2:
        config = NeighborhoodConfig(True, lambda x, i: x["levels"][i])
        generator = neighborhood2or3(graph1, graph2, mapping, cmp_node, config)

    if k == 3:
        config = NeighborhoodConfig(False, lambda x, i: x["node_indexes"])
        generator = neighborhood2or3(graph1, graph2, mapping, cmp_node, config)

    for match in generator:
        yield match

    yield mapping


def vnd(graph1, graph2, neighborhoods=3, time_limit=0):
    """Calculate graph matching through VND algorithm"""
    if time_limit == 0:
        time_limit = 315569000
    time_limit *= 1000
    nodes1, nodes2 = prepare_graph(graph1, graph2)
    sim = Similarity(graph1, graph2)
    start = time.time()
    first = first_solution(nodes1, nodes2)
    best = (first, sim(first))
    k = 1
    while k <= neighborhoods and (time.time() - start) < time_limit:
        try:
            current = max(
                ((n, sim(n)) for n in neighborhood(k, graph1, graph2, best[0])
                 if (time.time() - start) < time_limit),
                key=lambda x: x[1])
        except ValueError:
            current = best
        if current[1] > best[1]:
            best = current
            k = 1
        else:
            k += 1
    return MappingToGraph(graph1, graph2, best[0]).to_dict()


def trial_mirror(func):
    """Decorator: Compute graphs for each trial before diff"""
    name = func.__name__

    def calculate(self, **kwargs):
        """Compute graphs for each trial before diff"""
        finished1, graph1 = getattr(self.diff.trial1.graph, name)()
        finished2, graph2 = getattr(self.diff.trial2.graph, name)()
        finished = finished1 and finished2
        return func(self, graph1, graph2, finished, **kwargs)
    return calculate


def class_param(params):
    """Decorator: Add extra params for cache"""
    params = params.split()

    def dec(func):
        """Decorator: Add extra params for cache"""
        def calculate(self, *args, **kwargs):
            """Add extra params for cache"""
            for param in params:
                value = getattr(self, param)
                if value is not None and param not in kwargs:
                    kwargs[param] = value
            return func(self, *args, **kwargs)
        return calculate
    return dec


cache = prepare_cache(                                                           # pylint: disable=invalid-name
    lambda self, *args, **kwargs: "diff {}:{}".format(self.diff.trial1.id,
                                                      self.diff.trial2.id))


class DiffGraph(Graph):                                                          # pylint: disable=too-many-instance-attributes
    """Diff Graph Class. Present diff graph on Jupyter"""

    def __init__(self, diff):
        self.diff = weakref.proxy(diff)

        self.use_cache = True
        self.width = 500
        self.height = 500
        self.mode = 3
        self.view = 0

        self._modes = {
            0: self.tree,
            1: self.no_match,
            2: self.exact_match,
            3: self.namespace_match
        }

        self._views = {
            0: "combined",
            1: "side by side",
            2: "both",
        }

        self.neighborhoods = None
        self.time_limit = None

    @class_param("neighborhoods time_limit")
    @cache("tree", "neighborhoods time_limit")
    @trial_mirror
    def tree(self, graph1, graph2, finished, **kwargs):                          # pylint: disable=no-self-use
        """Compare tree structures"""
        return finished, vnd(graph1, graph2, **kwargs), graph1, graph2

    @class_param("neighborhoods time_limit")
    @cache("no_match", "neighborhoods time_limit")
    @trial_mirror
    def no_match(self, graph1, graph2, finished, **kwargs):                      # pylint: disable=no-self-use
        """Compare graphs without matches"""
        return finished, vnd(graph1, graph2, **kwargs), graph1, graph2

    @class_param("neighborhoods time_limit")
    @cache("exact_match", "neighborhoods time_limit")
    @trial_mirror
    def exact_match(self, graph1, graph2, finished, **kwargs):                   # pylint: disable=no-self-use
        """Compare graphs with call matches"""
        return finished, vnd(graph1, graph2, **kwargs), graph1, graph2

    @class_param("neighborhoods time_limit")
    @cache("combine", "neighborhoods time_limit")
    @trial_mirror
    def namespace_match(self, graph1, graph2, finished, **kwargs):               # pylint: disable=no-self-use
        """Compare graphs with namespaces"""
        return finished, vnd(graph1, graph2, **kwargs), graph1, graph2

    def _repr_html_(self):
        """Display d3 graph on ipython notebook"""
        uid = str(int(time.time() * 1000000))

        return """
            <div class="nowip-diff" data-width="{width}"
                 data-height="{height}" data-uid="{uid}"
                 data-id1="{id1}" data-id2="{id2}" data-mode="{view}">
                {data}
            </div>
        """.format(
            uid=uid, id1=self.diff.trial1.id, id2=self.diff.trial2.id,
            view=self.view,
            data=self.escape_json(self._modes[self.mode]()[1:]),
            width=self.width, height=self.height
        )



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""File Access Model"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os

from sqlalchemy import Column, Integer, Text, TIMESTAMP
from sqlalchemy import PrimaryKeyConstraint, ForeignKeyConstraint

from ...utils.prolog import PrologDescription, PrologTrial, PrologAttribute
from ...utils.prolog import PrologRepr, PrologTimestamp, PrologNullable
from ...utils.prolog import PrologNullableRepr

from .. import relational, persistence_config

from .base import AlchemyProxy, proxy_class, backref_one, proxy


@proxy_class
class FileAccess(AlchemyProxy):
    """Represent a file access"""

    __tablename__ = "file_access"
    __table_args__ = (
        PrimaryKeyConstraint("trial_id", "id"),
        ForeignKeyConstraint(["trial_id", "function_activation_id"],
                             ["function_activation.trial_id",
                              "function_activation.id"], ondelete="CASCADE"),
        ForeignKeyConstraint(["trial_id"], ["trial.id"], ondelete="CASCADE"),
    )
    trial_id = Column(Integer, index=True)
    id = Column(Integer, index=True)                                             # pylint: disable=invalid-name
    name = Column(Text)
    mode = Column(Text)
    buffering = Column(Text)
    content_hash_before = Column(Text)
    content_hash_after = Column(Text)
    timestamp = Column(TIMESTAMP)
    function_activation_id = Column(Integer, index=True)

    trial = backref_one("trial")  # Trial.file_accesses
    activation = backref_one("activation")  # Activation.file_accesses

    prolog_description = PrologDescription("access", (
        PrologTrial("trial_id", link="activation.trial_id"),
        PrologAttribute("id", fn=lambda obj: "f{}".format(obj.id)),
        PrologRepr("name"),
        PrologRepr("mode"),
        PrologNullableRepr("content_hash_before"),
        PrologNullableRepr("content_hash_after"),
        PrologTimestamp("timestamp"),
        PrologNullable("activation_id", attr_name="function_activation_id",
                       link="activation.id"),
    ), description=(
        "informs that in a given trial (*trial_id*),\n"
        "a file *name*\n"
        "was accessed by a function activation (*activation_id*)\n"
        "in a specific read or write *mode*\n"
        "at a specific *timestamp*.\n"
        "The content of the file\n"
        "was captured before (*content_hash_before*)\n"
        "and after (*content_hash_after*) the access."
    ))

    @property
    def stack(self):
        """Return the activation stack since the beginning of execution"""
        stack = []
        activation = self.activation
        while activation:
            name = activation.name
            activation = activation.caller
            if activation:
                stack.insert(0, name)
        if not stack or stack[-1] != "open":
            stack.append(" ... -> open")
        return " -> ".join(stack)

    @property
    def brief(self):
        """Brief description of file access"""
        result = "({0.mode}) {0.name}".format(self)
        if self.content_hash_before is None:
            result += " (new)"
        return result

    @property
    def activation_id(self):
        return self.function_activation_id

    @property
    def is_internal(self):
        return (
            not os.path.isabs(self.name) or
            persistence_config.base_path in self.name
        )

    @classmethod  # query
    def find_by_name_and_time(cls, name, timestamp, trial=None,
                              session=None):
        """Return the first access according to name and timestamp

        Arguments:
        name -- specify the desired file
        timestamp -- specify the start of finish time of trial

        Keyword Arguments:
        trial -- limit search in a specific trial_id
        """
        model = cls.m
        session = session or relational.session
        query = (
            session.query(model)
            .filter(
                (model.name == name) &
                (model.timestamp.like(timestamp + "%"))
            ).order_by(model.timestamp)
        )
        if trial:
            query = query.filter(model.trial_id == trial)
        return proxy(query.first())

    def __key(self):
        return (self.name, self.content_hash_before, self.content_hash_after,
                self.mode)

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        if not isinstance(other, FileAccess):
            return False
        return (
            (self.content_hash_before == other.content_hash_before)
            and (self.content_hash_after == other.content_hash_after)
        )

    def show(self, _print=print):
        """Show object

        Keyword arguments:
        _print -- custom print function (default=print)
        """
        _print("""\
            Name: {f.name}
            Mode: {f.mode}
            Buffering: {f.buffering}
            Content hash before: {f.content_hash_before}
            Content hash after: {f.content_hash_after}
            Timestamp: {f.timestamp}
            Function: {f.stack}\
            """.format(f=self))

    def __repr__(self):
        return "FileAccess({0.trial_id}, {0.id}, {0.name}, {0.mode})".format(
            self)


class UniqueFileAccess(FileAccess):

    def __key(self):
        return self.id

    def __eq__(self, other):
        if not isinstance(other, FileAccess):
            return False
        return (self.id == other.id)

    def __hash__(self):
        return hash(self.__key())



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Tag Model"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from datetime import datetime

from future.utils import lmap
from future.builtins import map as cvmap
from sqlalchemy import Column, Integer, Text, TIMESTAMP
from sqlalchemy import ForeignKeyConstraint, select, bindparam

from ...utils.prolog import PrologDescription, PrologTrial, PrologAttribute
from ...utils.prolog import PrologRepr, PrologTimestamp

from .. import relational

from .base import AlchemyProxy, proxy_class, backref_one


@proxy_class
class Tag(AlchemyProxy):
    """Represent a tag"""

    __tablename__ = "tag"
    __table_args__ = (
        ForeignKeyConstraint(["trial_id"], ["trial.id"], ondelete="CASCADE"),
        {"sqlite_autoincrement": True},
    )
    id = Column(Integer, primary_key=True)                                       # pylint: disable=invalid-name
    trial_id = Column(Integer, index=True)
    type = Column(Text)                                                          # pylint: disable=invalid-name
    name = Column(Text)
    timestamp = Column(TIMESTAMP)

    trial = backref_one("trial")  # Trial.inherited

    prolog_description = PrologDescription("tag", (
        PrologTrial("trial_id", link="trial.id"),
        PrologRepr("type"),
        PrologRepr("name"),
        PrologTimestamp("timestamp"),
    ), description=(
        "informs that a given trial (*trial_id*)\n"
        "has a tag (*name*) of *type*,\n"
        "created at *timestamp*.\n"
    ))

    def __repr__(self):
        return "Tag({0.trial_id}, {0.type}, {0.name})".format(self)

    @classmethod  # query
    def fast_load_auto_tag(cls, trial_id, code_hash, command,
                           session=None):
        """Find automatic by code_hash and command.
        Ignore tags on the same trial_id


        Return (typ, tag)
        typ -- int representing the type of tag:
            0: Completely new tag (1.1.1)
            1: Match both code_hash and command (new tag should be x.y.+)
            2: Match code_hash (new tag should be x.+.1)
            3: New code_hash (new tag should be +.1.1)
        tag -- list with the found tag


        Arguments:
        trial_id -- id of trial that should be tagged
        code_hash -- code_hash of trial script
        command -- command line


        Keyword arguments:
        session -- specify session for loading (default=relational.session)
        """
        from .trial import Trial
        session = session or relational.session
        ttag = cls.__table__
        ttrial = Trial.__table__
        _query = select([ttag.c.name]).where(
            (ttrial.c.id == ttag.c.trial_id) &
            (ttrial.c.id != bindparam("trial_id")) &
            (ttag.c.type == "AUTO")
        )

        conditions = [
            (1, ((ttrial.c.code_hash == bindparam("code_hash")) &
                 (ttrial.c.command == bindparam("command")))),
            (2, ((ttrial.c.code_hash == bindparam("code_hash")))),
            (3, True)
        ]

        info = {
            "trial_id": trial_id,
            "code_hash": code_hash,
            "command": command,
        }

        for typ, condition in conditions:
            results = session.execute(_query.where(condition), info).fetchall()
            tags = [lmap(int, tag[0].split(".")) for tag in results]
            if tags:
                return typ, max(tags)

        return 0, [1, 1, 1]

    @classmethod  # query
    def create_automatic_tag(cls, trial_id, code_hash, command,
                             session=None):
        """Create automatic tag for trial id

        Find maximum automatic tag by code_hash and command
        If it has the same code_hash and command, increment patch
        If it has the same code_hash, increment minor
        Otherwise, increment major


        Arguments:
        trial_id -- id of trial that should be tagged
        code_hash -- code_hash of trial script
        command -- command line


        Keyword arguments:
        session -- specify session for loading (default=relational.session)
        """
        session = session or relational.session
        tag_typ, tag = cls.fast_load_auto_tag(
            trial_id, code_hash, command, session=session)
        new_tag = ""
        if tag_typ == 1:
            tag[2] += 1
        elif tag_typ == 2:
            tag[1] += 1
            tag[2] = 1
        elif tag_typ == 3:
            tag[0] += 1
            tag[1] = 1
            tag[2] = 1
        new_tag = ".".join(cvmap(str, tag))

        session.execute(cls.t.insert(), dict(
            trial_id=trial_id, type="AUTO",
            name=new_tag, timestamp=datetime.now()
        ))
        session.commit()



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""'%now_schema' magic"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from sqlalchemy.schema import CreateTable

from ...cmd.cmd_schema import Schema

from ...utils.formatter import PrettyLines
from ...persistence.models import ORDER
from ...persistence import relational

from .command import IpythonCommandMagic


class NowSchema(IpythonCommandMagic, Schema):
    """Present Schema

    Examples
    --------
    ::
        In [1]: %now_schema sql -t
        Out[1]: CREATE TABLE trial (
                ...

        In [2]: %now_schema prolog --format png
        Out[2]: <png graph>
    """

    def __init__(self, *args, **kwargs):
        super(NowSchema, self).__init__(*args, **kwargs)
        self.format = "svg"

    def add_arguments(self):
        super(NowSchema, self).add_arguments()
        add_arg = self.add_argument
        add_arg("-f", "--format", type=str.lower, default="svg",
                choices=["svg", "png"],
                help="Graph format for diagram export (default: svg)")
        add_arg("-t", "--text", action="store_true",
                help="export textual schema")

    def post_process(self, result, args):
        """Transform result into image or PrettyLines"""                         # pylint: disable=no-self-use
        if args.diagram:
            result.format = self.format
            return result
        return PrettyLines(result)

    def execute(self, func, line, cell, magic_cls):
        _, args = self.arguments(func, line)
        args.diagram = not args.text
        self.format = args.format
        return self.process(args)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Command base"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import argparse

from ..utils.functions import abstract


class SmartFormatter(argparse.HelpFormatter):
    """Add option to split lines in help messages"""

    def _split_lines(self, text, width):
        # this is the RawTextHelpFormatter._split_lines
        if text.startswith('R|'):
            return text[2:].splitlines()
        return argparse.HelpFormatter._split_lines(self, text, width)


class Command(object):
    """Command base"""

    def __init__(self, cmd=None):
        self.cmd = cmd or type(self).__name__.lower()
        self.help = self.__doc__.split("\n")[0].strip()
        self.parser = None
        self.add_help = True
        self.is_ipython = False

    def create_parser(self, subparsers):
        """Create parser with arguments"""
        kwargs = {}
        if self.help:
            kwargs["help"] = self.help
        self.parser = subparsers.add_parser(
            self.cmd, add_help=self.add_help, formatter_class=SmartFormatter,
            **kwargs
        )

        self.add_arguments()
        self.parser.set_defaults(func=self.execute)

    def add_arguments(self):
        """Add arguments to command. Override on subclass"""
        pass

    def add_argument(self, *args, **kwargs):
        """Add argument to parser available for both IPython magic and cmd"""
        return self.parser.add_argument(*args, **kwargs)

    def add_argument_cmd(self, *args, **kwargs):
        """Add argument to parser available only for cmd"""
        return self.parser.add_argument(*args, **kwargs)

    def execute(self, args):
        """Execute the command. Override on subclass"""
        abstract()
        print(self, args)


class NotebookCommand(Command):
    """NotebookCommand base. Default option -i to export notebook files"""

    def create_parser(self, subparsers):
        """Create parser with arguments"""
        super(NotebookCommand, self).create_parser(subparsers)
        self.add_argument("-i", "--ipynb", action="store_true",
                          help="export jupyter notebook file")
        self.parser.set_defaults(func=self._execute)

    def _execute(self, args):
        """Select between export or execute"""
        if args.ipynb:
            self.execute_export(args)
        else:
            self.execute(args)

    def execute_export(self, args):
        """Export notebook file. Override on subclass"""
        abstract()
        print(self, args)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Object Value Model"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

from sqlalchemy import Column, Integer, Text
from sqlalchemy import PrimaryKeyConstraint, ForeignKeyConstraint
from sqlalchemy import CheckConstraint

from ...utils.prolog import PrologDescription, PrologTrial, PrologAttribute
from ...utils.prolog import PrologRepr

from .base import AlchemyProxy, proxy_class, backref_one


@proxy_class
class ObjectValue(AlchemyProxy):
    """Represent an object value (global, argument)"""

    __tablename__ = "object_value"
    __table_args__ = (
        PrimaryKeyConstraint("trial_id", "function_activation_id", "id"),
        ForeignKeyConstraint(["trial_id"], ["trial.id"], ondelete="CASCADE"),
        ForeignKeyConstraint(["trial_id", "function_activation_id"],
                             ["function_activation.trial_id",
                              "function_activation.id"], ondelete="CASCADE"),
    )
    trial_id = Column(Integer, index=True)
    function_activation_id = Column(Integer, index=True)
    id = Column(Integer, index=True)                                             # pylint: disable=invalid-name
    name = Column(Text)
    value = Column(Text)
    type = Column(Text, CheckConstraint("type IN ('GLOBAL', 'ARGUMENT')"))       # pylint: disable=invalid-name

    trial = backref_one("trial")  # Trial.object_values
    activation = backref_one("activation")  # Ativation.object_values

    prolog_description = PrologDescription("object_value", (
        PrologTrial("trial_id", link="activation.trial_id"),
        PrologAttribute("activation_id", attr_name="function_activation_id",
                        link="activation.id"),
        PrologAttribute("id"),
        PrologRepr("name"),
        PrologRepr("value"),
        PrologRepr("type"),
    ), description=(
        "informs that in a given trial (*trial_id*),\n"
        "a given activation (*function_activation_id*),\n"
        "has a GLOBAL/ARGUMENT (*type*) variable *name*,\n"
        "with *value*.\n"
    ))

    def __repr__(self):
        return (
            "ObjectValue({0.trial_id}, {0.function_activation_id}, {0.id}, "
            "{0.name}, {0.value}, {0.type})"
        ).format(self)

    def __str__(self):
        return "{0.name} = {0.value}".format(self)



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""'now schema'command"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import os

from sqlalchemy.schema import CreateTable

from ..persistence.models.graphs.diagram import ViewRelationalDiagram
from ..persistence.models import Trial, TrialProlog, ORDER
from ..persistence import persistence_config, relational


from .command import Command


class Schema(Command):
    """Present the SQL or Prolog schema of noWorkflow"""

    def __init__(self, *args, **kwargs):
        super(Schema, self).__init__(*args, **kwargs)

    def add_arguments(self):
        add_arg = self.add_argument
        self.add_argument_cmd("-d", "--diagram", action="store_true",
                              help="export graphic schema to dot")
        add_arg("type", type=str.lower, choices=["sql", "prolog"],
                help="schema type")
        add_arg("--dir", type=str,
                help="set project path where is the database. Default to "
                     "current directory")

    def sql_text(self):                                                          # pylint: disable=no-self-use
        """Export database creation schema"""
        lines = []
        for model in ORDER:
            lines += (
                str(CreateTable(model.t)
                    .compile(relational.engine)).split("\n")
            )
        return lines

    def prolog_text(self):                                                       # pylint: disable=no-self-use
        """Export prolog rules and fact descriptions"""
        trial_prolog = TrialProlog(Trial())
        return trial_prolog.rules(with_facts=True)

    def sql_diagram(self):                                                       # pylint: disable=no-self-use
        """Export SQL diagram"""
        return ViewRelationalDiagram(ORDER)

    def prolog_diagram(self):                                                    # pylint: disable=no-self-use
        """Export Prolog diagram"""
        return TrialProlog.diagram()

    def post_process(self, result, args):
        """Transform result into text"""                                         # pylint: disable=no-self-use
        if args.diagram:
            return result.as_dot()
        return "\n".join(result)

    def process(self, args):
        """Process schema"""
        func = {
            "sql": [self.sql_text, self.sql_diagram],
            "prolog": [self.prolog_text, self.prolog_diagram],
        }[args.type][args.diagram]
        return self.post_process(func(), args)

    def execute(self, args):
        persistence_config.connect_existing(args.dir or os.getcwd())
        print(self.process(args))



# Copyright (c) 2016 Universidade Federal Fluminense (UFF)
# Copyright (c) 2016 Polytechnic Institute of New York University.
# This file is part of noWorkflow.
# Please, consult the license terms in the LICENSE file.
"""Tracer Provider. Perform program slicing"""
from __future__ import (absolute_import, print_function,
                        division, unicode_literals)

import sys
import linecache
import traceback

from collections import namedtuple
from datetime import datetime
from functools import partial
from inspect import ismethod
from copy import copy

from future.utils import viewitems

from ...persistence.models import Variable, VariableDependency
from ...persistence.models import VariableUsage
from ...utils.io import print_fn_msg
from ...utils.bytecode.f_trace import find_f_trace, get_f_trace
from ...utils.cross_version import IMMUTABLE, builtins
from ...utils.functions import NOWORKFLOW_DIR

from ..prov_definition.utils import CallDependency, Dependency
from ..prov_definition.utils import Variable as Var

from .argument_captors import SlicingArgumentCaptor
from .profiler import Profiler


ActivationSlicing = namedtuple("ActivationSlicing",
                               "call_var return_var activation_id id")


class JointPartial(partial):                                                     # pylint: disable=inherit-non-class, too-few-public-methods
    """Combine noWorkflow Tracer with debugger Tracer"""

    def __init__(self):
        super(JointPartial, self).__init__()
        self._joint_tracer = True

    def __repr__(self):
        return "Joint<{}>".format(", ".join("{}.{}".format(
            (a.im_self if ismethod(a) else a).__class__.__name__, a.__name__
            ) for a in self.args))

    @property
    def __name__(self):
        return "Joint"


def joint_tracer(first, second, frame, event, arg):
    """Joint tracer used to combine noWorkflow tracer with other tracers"""
    try:
        first = first(frame, event, arg)
    except Exception:                                                            # pylint: disable=broad-except
        traceback.print_exc()
    try:
        second = second(frame, event, arg)
    except Exception:                                                            # pylint: disable=broad-except
        traceback.print_exc()
    return create_joint_tracer(first, second)


def create_joint_tracer(first, second):
    """Create joint tracer"""
    if not first:
        return second
    if not second:
        return first
    if first == second:
        return first
    joint = partial(joint_tracer, first, second)
    joint._joint_tracer = True                                                   # pylint: disable=protected-access
    # joint.__str__ = lambda x: "Joint<{}, {}>".format(first, second)

    return joint

_sys_settrace = sys.settrace                                                     # pylint: disable=invalid-name


def joint_settrace(tracer):
    """Override sys.settrace to create joint tracer"""
    current = sys.gettrace()
    if hasattr(current, "_joint_tracer"):
        new = create_joint_tracer(current.args[0], tracer)
    elif current:
        new = create_joint_tracer(current, tracer)
    else:
        new = tracer
    _sys_settrace(new)

sys.settrace = joint_settrace


class ActivationLoop(object):
    """Activation Loop class. Used for tracking variables in For loops"""

    def __init__(self, loop):
        self.iterable = []
        self.iter_var = []
        self.loop_def = loop
        self.remove = False
        self.temp_context = {}

    def __contains__(self, line):
        """Check if line is in loop"""
        return self.loop_def.first_line <= line <= self.loop_def.last_line


class ActivationCondition(object):
    """Activation Condition class. Used for tracking variables in conditions"""

    def __init__(self, condition):
        self.test_var = []
        self.remove = False
        self.condition_def = condition

    def __contains__(self, line):
        """Check if line is in loop"""
        condition_def = self.condition_def
        return condition_def.first_line <= line <= condition_def.last_line


def last_valid(objects):
    """Return last object not marked for removal"""
    for obj in reversed(objects):
        if not obj.remove:
            return obj
    return None


class Tracer(Profiler):                                                          # pylint: disable=too-many-instance-attributes
    """Tracer used for program slicing"""

    def __init__(self, *args):
        super(Tracer, self).__init__(*args)
        definition = self.metascript.definition
        self.event_map["line"] = self.trace_line

        # Store slicing provenance
        self.variables = self.metascript.variables_store
        self.dependencies = self.metascript.variables_dependencies_store
        self.usages = self.metascript.usages_store

        # Useful maps
        # Map of dependencies by line
        self.line_dependencies = definition.line_dependencies
        # Map of name_refs by line
        self.line_usages = definition.line_usages
        # Map of calls by line and col
        self.call_by_col = definition.call_by_col
        # Map of calls by line and lasti
        self.call_by_lasti = definition.call_by_lasti
        # Map of with __enter__ by line and lasti
        self.with_enter_by_lasti = definition.with_enter_by_lasti
        # Map of with __exit__ by line and lasti
        self.with_exit_by_lasti = definition.with_exit_by_lasti

        # Set of imports
        self.imports = definition.imports
        # Set of GET_ITER and FOR_ITER lasti by line
        self.iters = definition.iters
        # Map of For and While loops
        self.loops = definition.loops
        # Map of If and While statement conditions
        self.conditions = definition.conditions

        # Allow debuggers:
        self.f_trace_frames = []

        # Events are not unique. Profiler and Tracer have same events
        self.unique_events = False

        # Created builtins
        self.created_builtins = {}
        self.builtins = builtins.__dict__

        # Blackbox
        self.blackbox = None
        self.blackbox_index = 1

        self.argument_captor = SlicingArgumentCaptor(self)

        # List of calls in comprehension
        self.comprehension_dependencies = None


    def add_variable(self, act_id, name, line, f_locals, typ, value="--chk--"):     # pylint: disable=too-many-arguments
        """Add variable


        Arguments:
        act_id -- Activation object id
        name -- Variable name
        line -- Variable line
        f_locals -- Local Variables Dict in variable frame

        Keyword argument:
        value -- override variable value (default "--chk--")
        """
        if value == "--chk--" and name in f_locals:
            value = self.serialize(f_locals[name])
        else:
            value = "now(n/a)"
        return self.variables.add(
            act_id, name, line, value, datetime.now(), typ)


    def find_variable(self, activation, name, definition):
        """Find variable in activation context by name


        Arguments:
        activation -- current activation
        name -- variable name or tuple
        definition -- definition filename
        """
        name = get_call_var(name)

        if name in activation.context:
            # Local context
            return get_call_var(activation.context[name])
        elif name in self.globals[definition].context:
            # Global contex
            activation = self.globals[definition]
            return get_call_var(activation.context[name])
        elif name in self.created_builtins:
            # Buitins
            activation = self.main_activation
            return get_call_var(self.created_builtins[name])
        elif name in self.builtins:
            # New Builtins
            activation = self.main_activation
            vid = self.add_variable(activation.id, name, 0, self.builtins,
                                    "builtin")
            variable = self.variables[vid]
            self.created_builtins[name] = variable
            return get_call_var(variable)

    def find_variables(self, activation, dependencies, filename):
        """Find variables in activation context by names


        Arguments:
        activation -- current activation
        name -- variable name or tuple
        definition -- definition filename
        """
        for dep in dependencies:
            variable = self.find_variable(activation, dep.dependency, filename)
            if variable is None and isinstance(dep.dependency, tuple):
                variable = self.add_fake_call(activation, dep.dependency)
            if variable:
                yield variable, dep.type
            if dep.type == "loop":
                for loop in activation.loops:
                    for variable in loop.iter_var:
                        yield variable, "loop"

    def add_dependencies(self, dependent, dependencies, replace=None):
        """ Create dependencies: dependent depends on dependencies

        Arguments:
        self -- Tracer object
        dependent -- Dependent Variable object
        dependencies -- List of tuples (dependency variable, dependency type)
        """
        dependencies_add = self.dependencies.add
        for target, dep_type in dependencies:
            if replace:
                dep_type = replace
            dependencies_add(dependent.activation_id, dependent.id,
                             target.activation_id, target.id, dep_type)

    def slice_dependencies(self, activation, lineno, f_locals, var, deps):      # pylint: disable=too-many-arguments
        """Create dependencies for variable"""
        vid = None
        if isinstance(var, Var):
            value = "--chk--"
            if var.name == "return":
                value = activation.return_value
            vid = self.add_variable(activation.id, var.name, lineno, f_locals,
                                    var.type, value=value)

        if vid is not None:
            variable = self.variables[vid]
            self.add_dependencies(variable, deps)
            activation.context[var.name] = variable
            if var == "yield":
                activation.context["return"] = activation.context[var.name]
            for condition in activation.conditions:
                self.add_dependencies(variable, condition.test_var)
            for condition in activation.permanent_conditions:
                self.add_dependencies(variable, condition.test_var)
        elif var in activation.context:
            # var is a tuple representing a call.
            if isinstance(var, CallDependency):
                variable = activation.context[var].call_var
            else:
                variable = activation.context[var].return_var
            self.add_dependencies(variable, deps)
        else:
            # Python skips tracing calls to builtins (float, int...)
            # create artificial variables for them
            variable = self.add_fake_call(activation, var)
            self.add_dependencies(variable, deps)

        return variable

    def slice_loop(self, activation, lineno, f_locals, filename):
        """Create loops, and generates dependencies between iterables"""
        loops = activation.loops
        while loops and lineno not in loops[-1]:
            loops.pop()
        context = activation.context
        loop_def = self.loops[filename].get(lineno)
        if loop_def is not None and loop_def.first_line == lineno:
            if not loops or loops[-1].loop_def != loop_def:
                loop = ActivationLoop(loop_def)

                loop.iterable = list(self.find_variables(
                    activation, loop_def.iterable, filename))

                loop.iter_var = []
                loop.first_iter = True

                loops.append(loop)
        if loops and loops[-1].loop_def.first_line_in_scope == lineno:
            loop = loops[-1]
            loop_def = loop.loop_def
            loop.iter_var = []
            for var in loop_def.iter_var:
                loop.iter_var.append(self.slice_dependencies(
                    activation, loop_def.first_line, f_locals, var,
                    loop.iterable))

            if loop.first_iter:
                for var in activation.temp_context:
                    loop.temp_context[var] = context[var]
                loop.first_iter = False

            for var_name, var in viewitems(loop.temp_context):
                activation.temp_context.add(var_name)
                activation.context[var_name] = var

    def slice_condition(self, activation, lineno, f_locals, filename):           # pylint: disable=unused-argument
        """Create if and while conditions"""
        conditions = activation.conditions
        while conditions and lineno not in conditions[-1]:
            condition = conditions.pop()
            if condition.condition_def.has_return:
                activation.permanent_conditions.append(condition)

        condition_def = self.conditions[filename].get(lineno)
        if condition_def is not None and condition_def.first_line == lineno:
            if not conditions or conditions[-1].condition_def != condition_def:
                condition = ActivationCondition(condition_def)

                condition.test_var = list(self.find_variables(
                    activation, condition_def.test_var, filename))
                conditions.append(condition)

            elif conditions[-1].condition_def == condition_def:
                condition = conditions[-1]

                condition.test_var = list(self.find_variables(
                    activation, condition_def.test_var, filename))

    def slice_line(self, activation, lineno, f_locals, filename):
        """Generates dependencies from line"""
        for var in activation.temp_context:
            del activation.context[var]
        activation.temp_context = set()

        print_fn_msg(lambda: "Slice [{}] -> {}".format(
            lineno, linecache.getline(filename, lineno).strip()))

        self.slice_loop(activation, lineno, f_locals, filename)
        self.slice_condition(activation, lineno, f_locals, filename)

        context = activation.context
        usages_add = self.usages.add

        for ctx in ("Load", "Del"):
            usages = self.line_usages[filename][lineno][ctx]
            for name in usages:
                if name in context:
                    usages_add(activation.id, context[name].id, lineno, ctx)

        for var, others in viewitems(self.line_dependencies[filename][lineno]):
            deps = self.find_variables(activation, others, filename)
            #deps = list(deps)
            self.slice_dependencies(activation, lineno, f_locals, var, deps)

    def add_fake_call(self, activation, call_uid):
        """Create fake call for builtins"""
        line, col = call_uid
        call = self.call_by_col[activation.definition_file][line][col]
        vid = self.add_variable(activation.id, call.name,
                                line, {}, call.prefix, value="now(n/a)")
        variable = self.variables[vid]

        if "import" in call.name:
            box = self.create_blackbox()
        else:
            box = self.create_graybox()

        self.dependencies.add(variable.activation_id, variable.id,
                              box.activation_id, box.id, "box")

        activation.context[call_uid] = ActivationSlicing(
            variable, variable, activation.id, variable.id)
        activation.temp_context.add(call_uid)

        args = self.find_variables(activation, call.all_args(),
                                   activation.definition_file)

        self.add_dependencies(box, args)
        return variable

    def create_call(self, activation, _return):
        """Create call variable and dependency"""
        caller = self.current_activation
        dependencies_add = self.dependencies.add
        typ = "call"
        if activation.name in ("_handle_fromlist", "_find_and_load"):
            typ = "import"
        vid = self.add_variable(caller.id, activation.name,
                                activation.line, {}, typ, value="now(n/a)")
        dependencies_add(caller.id, vid, activation.id, _return.id,
                         "return")

        if caller.with_definition:
            filename = activation.filename
            line, lasti = activation.line, activation.lasti
            try:
                call = self.call_by_lasti[filename][line][lasti]
            except (IndexError, KeyError):
                call = None
                # call not found
                # ToDo: show in dev-mode
            if call is not None:
                uid = (call.line, call.col)
                if uid in caller.context and line in self.imports[filename]:
                    # Two calls in the same lasti: create dependencies import
                    variable = caller.context[uid].call_var
                    dependencies_add(activation.id, _return.id,
                                     variable.activation_id, variable.id,
                                     "call")
                caller.context[uid] = ActivationSlicing(
                    self.variables[vid], _return, caller.id, vid)


        if self.comprehension_dependencies is not None:
            if activation.is_comprehension():
                for dependency in self.comprehension_dependencies:
                    dependencies_add(activation.id, _return.id,
                                     dependency[0], dependency[1], "direct")
                self.comprehension_dependencies = None
            else:
                self.comprehension_dependencies.append((caller.id, vid))

    def close_activation(self, frame, event, arg):
        """Slice all lines from closing activation
        Create Call variable and Add generic return if frame is not None
        """
        activation = self.current_activation
        for line in activation.slice_stack:
            self.slice_line(*line)
        super(Tracer, self).close_activation(frame, event, arg)
        if frame and not activation.is_main:
            _return = self.add_generic_return(activation, frame)
            _return.value = activation.return_value
            self.create_call(activation, _return)

    def create_blackbox(self):
        """Create a blackbox object with dependency to the previous one"""
        vid = self.add_variable(0, "--blackbox--", self.blackbox_index,
                                {}, "--blackbox--", value="now(n/a)")
        blackbox = self.variables[vid]
        old_blackbox = self.blackbox
        if old_blackbox is not None:
            self.dependencies.add(blackbox.activation_id, blackbox.id,
                                  old_blackbox.activation_id, old_blackbox.id,
                                  "box")
        self.blackbox = blackbox
        self.blackbox_index += 1
        return blackbox

    def create_graybox(self):
        """Create a graybox object"""
        vid = self.add_variable(0, "--graybox--", 0,
                                {}, "--graybox--", value="now(n/a)")
        return self.variables[vid]


    def add_generic_return(self, activation, frame):
        """Add return to functions that do not have return
        For ccall, add dependency from all params


        Arguments:
        activation -- closing Activation
        frame -- parent Frame
        """
        lineno = frame.f_lineno
        variables = self.variables

        # Artificial return condition
        if "return" in activation.context:
            return activation.context["return"] # call has return

        vid = self.add_variable(activation.id, "return", lineno, {},
                                "virtual",
                                value=activation.return_value)
        _return = variables[vid]

        activation.context["return"] = _return

        if activation.with_definition:
            # we captured the slicing already
            # return does not depend on parameters
            return _return

        # Artificial return depends on blackbox
        definition_file = activation.definition_file
        if NOWORKFLOW_DIR in definition_file or definition_file == "now(n/a)":
            # noworkflow activations have no side effect
            # ToDo: check if it is builtin or just C
            blackbox = self.create_graybox()
        else:
            blackbox = self.create_blackbox()
        self.dependencies.add(_return.activation_id, _return.id,
                              blackbox.activation_id, blackbox.id, "box")

        if not activation.has_parameters:
            # activation does not have parameters
            return _return

        caller = self.current_activation
        if not caller.with_definition:
            # we do not have caller definition
            # thus, we do not know activation parameters
            return _return

        # Artificial return depends on all parameters
        filename = activation.filename
        line, lasti = activation.line, activation.lasti
        try:
            call = self.call_by_lasti[filename][line][lasti]
        except (IndexError, KeyError):
            # call not found
            # ToDo: show in dev-mode
            return _return

        all_args = list(self.find_variables(caller, call.all_args(), filename))
        self.add_dependencies(blackbox, all_args)
        self.add_inter_dependencies(frame.f_locals, all_args, caller, line,
                                    [(blackbox, "box")])
        return _return

    def add_inter_dependencies(self, f_locals, args, caller, lineno, other):     # pylint: disable=too-many-locals
        """Add dependencies between all parameters in a call


        Arguments:
        f_locals -- Frame local variables
        args -- Parameter Variables
        caller -- parent Activation
        lineno -- Activation lineno
        other -- Other args
        """
        context = caller.context
        variables = self.variables
        add_dependencies = self.add_dependencies
        add_variable = self.add_variable

        added = {}
        for arg, dep_type in args:
            try:
                name = arg.name
                var = f_locals[name]
                if not isinstance(var, IMMUTABLE):
                    vid = add_variable(caller.id, name, lineno, {},
                                       "arg", value=var)
                    variable = variables[vid]
                    add_dependencies(variable, other, replace="argument")
                    added[name] = variable
                #else:
                #    arg.value = self.serialize(var)
            except KeyError:
                pass

        for name, variable in viewitems(added):
            context[name] = variable

    def trace_line(self, frame, event, arg):                                     # pylint: disable=unused-argument
        """Trace Line event"""
        code = frame.f_code
        filename = frame.f_code.co_filename
        lineno = frame.f_lineno

        loc, glob = frame.f_locals, frame.f_globals
        if find_f_trace(code, loc, glob, frame.f_lasti):
            _frame = get_f_trace(code, loc, glob)
            if _frame.f_trace:
                self.f_trace_frames.append(_frame)

        activation = self.current_activation

        if not activation.with_definition:
            return  # different file

        if activation.is_comprehension():
            self.comprehension_dependencies = []
            return  # ignore comprehension

        last_loop = last_valid(activation.loops)
        if last_loop and lineno not in last_loop:
            # Remove last trace line
            activation.slice_stack.pop()
            last_loop.remove = True

        print_fn_msg(lambda: "[{}] -> {}".format(
            lineno, linecache.getline(filename, lineno).strip()))
        if activation.slice_stack:
            self.slice_line(*activation.slice_stack.pop())
        activation.slice_stack.append([
            activation, lineno, frame.f_locals, filename])

    def trace_pre_tracer(self, frame, event, arg):                               # pylint: disable=unused-argument
        """It is executed before the tracing event. Check f_trace is set"""
        self.check_f_trace(frame, event)

    def check_f_trace(self, frame, event):                                       # pylint: disable=unused-argument
        """Check if frame changes current tracer
        Replace the new tracer to a joint tracer
        """
        if self.f_trace_frames:

            for _frame in self.f_trace_frames:
                _frame.f_trace = create_joint_tracer(_frame.f_trace,
                                                     self.tracer)

            self.f_trace_frames = []
            return

    def tearup(self):
        """Activate tracer"""
        _sys_settrace(self.tracer)
        sys.setprofile(self.tracer)

    def teardown(self):
        """Deactivate tracer"""
        super(Tracer, self).teardown()
        _sys_settrace(self.default_trace)

    def store(self, partial=False):                                              # pylint: disable=redefined-outer-name
        """Store provenance"""
        if not partial:
            while len(self.activation_stack) > 1:
                self.close_activation(None, "store", None)
        super(Tracer, self).store(partial=partial)
        tid = self.trial_id
        Variable.fast_store(tid, self.variables, partial)
        VariableDependency.fast_store(tid, self.dependencies, partial)
        VariableUsage.fast_store(tid, self.usages, partial)

    def view_slicing_data(self, show=True):
        """View captured slicing"""
        if show:
            for var in self.variables:
                print_fn_msg(lambda v=var: v)

            for dep in self.dependencies:
                print_fn_msg(lambda d=dep: "{}\t<-\t{}".format(
                    self.variables[d.source_id],
                    self.variables[d.target_id]))

            for var in self.usages:
                print_fn_msg(lambda: var)


def get_call_var(name):
    """Return Call Var if is instance of ActivationSlicing"""
    if isinstance(name, ActivationSlicing):
        return name.call_var
    return name



#!/usr/bin/env python
from ez_setup import use_setuptools
use_setuptools()

from setuptools import setup, find_packages
from setuptools.command.install import install
from setuptools.command.develop import develop

import os
import platform

from uuid import getnode

with open("./noworkflow/resources/version.txt", "r") as f:
    __version__ = f.read().strip()


def recursive_path(pack, path):
    matches = []
    for root, dirnames, filenames in os.walk(os.path.join(pack, path)):
        for filename in filenames:
            matches.append(os.path.join(root, filename)[len(pack) + 1:])
    return matches


try:
    import pypandoc
    long_description = pypandoc.convert("../README.md", "rst")
except (IOError, ImportError):
    long_description = (
        "Supporting infrastructure to run scientific experiments "
        "without a scientific workflow management system.")


def analytics(command_subclass):
    """Sends install statistics to analytics."""
    orig_run = command_subclass.run

    def modified_run(self):
        import sys
        try:
            try:
                from urllib2 import urlopen, Request
                from urllib import urlencode
            except ImportError:
                from urllib.request import urlopen, Request
                from urllib.parse import urlencode

            os_ver = platform.system()
            py_ver = "_".join(str(x) for x in sys.version_info)
            now_ver = __version__.replace(".", "_")

            code = "os:{0},py:{1},now:{2}".format(os_ver, py_ver, now_ver)
            action = command_subclass.action
            cid = getnode()
            payload = {
                "v": "1",
                "tid": "UA-61791314-1",
                "cid": str(cid),
                "t": "event",
                "ec": action,
                "ea": code,
            }

            url = "http://www.google-analytics.com/collect"
            data = urlencode(payload).encode("utf-8")
            request = Request(url, data=data)
            request.get_method = lambda: "POST"
            urlopen(request)
        except:
            pass
        orig_run(self)

    command_subclass.run = modified_run
    return command_subclass


@analytics
class CustomDevelopCommand(develop):
    action = "develop"


@analytics
class CustomInstallCommand(install):
    action = "install"


setup(
    name="noworkflow",
    version=__version__,
    packages=find_packages(),
    package_data={
        "noworkflow": (
            recursive_path("noworkflow", "resources")
            + recursive_path("noworkflow", "now/vis/static")
            + recursive_path("noworkflow", "now/vis/templates")
        ),
    },
    entry_points={
        "console_scripts": ["now=noworkflow:main"]
    },
    author=("Joao Pimentel, Leonardo Murta, Vanessa Braganholo, "
            "Fernando Chirigati, David Koop, and Juliana Freire"),
    author_email="leomurta@ic.uff.br",
    description="Supporting infrastructure to run scientific experiments "
                "without a scientific workflow management system.",
    long_description=long_description,
    license="MIT",
    keywords="scientific experiments provenance python",
    url="https://github.com/gems-uff/noworkflow",
    install_requires=["pyposast>=1.1.3", "future", "SQLAlchemy"],
    extras_require={
        "vis": ["pyposast", "flask"],
        "notebook": ["pyposast", "ipython", "jupyter"],
        "all": ["pyposast", "ipython", "jupyter", "flask", "pyswip-alt",
                "jsonpickle"],
    },
    cmdclass={
        "install": CustomInstallCommand,
        "develop": CustomDevelopCommand,
    },
    classifiers=[
        'Development Status :: 5 - Production/Stable',

        'Intended Audience :: Science/Research',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',

        'License :: OSI Approved :: MIT License',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ]
)
